/**
 * Copyright (c) 2016 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "app_error.h"
#include "app_gpiote.h"
#include "app_pwm.h"
#include "app_timer.h"
#include "app_uart.h"
#include "app_util.h"
#include "app_util_platform.h"
#include "ble.h"
#include "ble_db_discovery.h"
#include "ble_gap.h"
#include "ble_hci.h"
#include "ble_nus_c.h"
#include "boards.h"
#include "bsp.h"
#include "bsp_btn_ble.h"
#include "fds.h"
#include "low_power_pwm.h"
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_scan.h"
#include "nrf_cli.h"
#include "nrf_cli_uart.h"
#include "nrf_delay.h"
#include "nrf_drv_clock.h"
#include "nrf_drv_gpiote.h"
#include "nrf_drv_ppi.h"
#include "nrf_drv_pwm.h"
#include "nrf_drv_saadc.h"
#include "nrf_drv_timer.h"
#include "nrf_drv_uart.h"
#include "nrf_gpio.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_log_internal.h"
#include "nrf_nvmc.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_queue.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "nrfx_timer.h"
#include "nrfx_pwm.h"
#include "sdk_errors.h"
#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "nrf_soc.h"
#include "nrf_fstorage.h"
#include "nrf_fstorage_sd.h"

#define NRF_LOG_MODULE_NAME app
#define NRF_CLI_ENABLED 1

#define APP_BLE_CONN_CFG_TAG 1  /**< Tag that refers to the BLE stack configuration set with @ref sd_ble_cfg_set. The default tag is @ref BLE_CONN_CFG_TAG_DEFAULT. */
#define APP_BLE_OBSERVER_PRIO 3 /**< BLE observer priority of the application. There is no need to modify this value. */

#define UART_TX_BUF_SIZE 256 /**< UART TX buffer size. */
#define UART_RX_BUF_SIZE 256 /**< UART RX buffer size. */

#define NUS_SERVICE_UUID_TYPE BLE_UUID_TYPE_VENDOR_BEGIN /**< UUID type for the Nordic UART Service (vendor specific). */

/******************************* DFH *************************/
#define ECHOBACK_BLE_UART_DATA 0 /**< Echo the UART data that is received over the Nordic UART Service (NUS) back to the sender. */
/******************************* DFH *************************/

#define Board_PS false       // Change to true for Perfect Shower Valve build, false for nRF52DK development board
#define Board_52DK !Board_PS // Used for debugging on nRF52DK system

// Pin Definitions for Perfect Shower RevD (MS50SFB) PCBA
#if Board_PS
// nRF52832 GPIO#                 // 52832 Pin# - Port0 # - MS50SFB Pin#
#define STEP_A 28        // pin #40 - P0_28/AIN4 - 6
#define STEP_B 29        // pin #41 - P0_29/AIN5 - 7
#define LED_1 17         // pin #20 - P0.17      - 32 Grn
#define LED_2 18         // pin #21 - P0.18      - 33 Red
#define SYS_V 30         // pin #42 - P0.30/AIN6 - 8
#define A_TEMP_OUT 31    // pin #43 - P0.31/AIN7 - 9
#define A_TEMP_IN 2      // pin #4  - P0.02/AIN0 - 16
#define AUDIO_OUT 3      // pin #5  - P0.03/AIN1 - 17 (buzzer)
#define MAN_OVERRIDE 11  // pin #14 - P0.11      - 26
#define VALVE_OFF 12     // pin #15 - P0.12      - 27
#define MAN_RESET 21     // pin #24 - P0.21/RST  - 36
#define TEMP_OVERRIDE 22 // pin #27 - P0.22      - 39
#define D_TEMP 23        // pin #28 - P0.23      - 40
#define ONEWIRE_PIN 23   // def for onewire code
#define TRIP_TEST 24     // pin #29 - P0.24      - 41
#define d_temp_irq 25    // pin #37 - P0.25      - P4
#endif

/* Pin Definitions for nRF52DK system */
/* nRF52DK GPIO#                  // 52832 Pin # - Port0.# - DK Hdr */
#if Board_52DK
#define LED_1 17         // pin #20 - P0.17 LED_1 - P3 Grn
#define LED_2 18         // pin #21 - P0.18 LED_2 - P3 Grn
#define STEP_A 28        // pin #40 - P0_28/AIN4  - P2   // Eric was 28
#define STEP_B 29        // pin #41 - P0_29/AIN5  - P2
#define SYS_V 30         // pin #42 - P0.30/AIN6  - P2
#define A_TEMP_OUT 31    // pin #43 - P0.31/AIN7  - P2
#define A_TEMP_IN 2      // pin #4  - P0.02/AIN0  - P4
#define AUDIO_OUT 3      // pin #5  - P0.03/AIN1  - P2 (buzzer)
#define MAN_OVERRIDE 11  // pin #14 - P0.11       - P3
#define VALVE_OFF 12     // pin #15 - P0.12       - P3
#define MAN_RESET 21     // pin #24 - P0.21/RST   - P6
#define TEMP_OVERRIDE 22 // pin #27 - P0.22       - P4
#define D_TEMP 23        // pin #28 - P0.23       - P4
#define ONEWIRE_PIN 23   // def for onewire code         // Eric was 23
#define TRIP_TEST 24     // pin #29 - P0.24       - P4
#define D_TEMP_IRQ 25    // pin 37  - P0.25       - P4
#endif
/*
// Unused pins
#define P0_26           26        // pin #38
#define P0_27           27        // pin #39
#define P0_00_XAL1      0         // pin #2
#define P0_01_XAL2      1         // pin #3
#define P0_04           4         // pin #6
#define P0_05_AIN3      5         // pin #7
#define P0_06           6         // pin #8
#define P0_07           7         // pin #9
#define P0_08           8         // pin #10
#define P0_09_NFC1      9         // pin #11
#define P0_10_NFC2      10        // pin #12
#define P0_11           11        // pin #14
#define P0_13           13        // pin #16
#define P0_14           14        // pin #17
#define P0_15           15        // pin #18
#define P0_16           16        // pin #19
#define P0_19           19        // pin #22
#define P0_20           20        // pin #23
*/

#define LEDS_NUMBER 2
#define LEDS_ACTIVE_STATE 0

#define LEDS_INV_MASK LEDS_MASK
#define LEDS_LIST { LED_1, LED_2 }
#define BSP_LED_0 LED_1
#define BSP_LED_1 LED_2

#define RX_PIN_NUMBER 8
#define TX_PIN_NUMBER 6
#define CTS_PIN_NUMBER 7
#define RTS_PIN_NUMBER 5
#define HWFC true
#define APP_SOC_OBSERVER_PRIO     1                                /**< Applications' SoC observer priority. You shouldn't need to modify this value. */

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

#define STARTUP_TIME 3000       /**< (app_timer_ticks) startup timer */
//#define STEPPER_MOTOR_RATE 10   /**< (app_timer_ticks) Stepper motor step rate between steps */
                                /**< between 5ms to 10ms optimum.  1ms too short) */

#define ECHOBACK_BLE_UART_DATA 0 /**< Echo the UART data that is received over the Nordic UART Service (NUS) back to the sender. */
#define SETSDATA(A, B, C, D, E, F, G, H, I) \
  sdata[0] = A;                    \
  sdata[1] = B;                    \
  sdata[2] = C;                    \
  sdata[3] = D;                    \
  sdata[4] = E;                    \
  sdata[5] = F;                    \
  sdata[6] = G;                    \
  sdata[7] = H;                    \
  sdata[8] = I;

typedef enum { FALSE = 0, TRUE = 1, } BOOL;
#define STEPPER_STATE_0(B, A), !STEP_B = B, !STEP_A = A;
//#define STEPPER_STATE_1(STEP_B,!STEP_A);
//#define STEPPER_STATE_2(STEP_B,STEP_A);
//#define STEPPER_STATE_3(!STEP_B,STEP_A);

static volatile int32_t stepper_motor_rate = 10;
static bool temp_scale = false;               // F/C. False=Farenheit, True=Celsius.  This value from user display
volatile static uint32_t current_temp;        // Global current water temperature
volatile static uint32_t t_raw;               // raw DS18B20 temperature sensor data
static  uint8_t t_count = 0;                 // temperature reading counter
volatile static uint32_t desired_temp = 40;   // This is the temperature desired by the User via the Wireless controller
static volatile uint32_t tdata[5];            // temperature storage for calculating average temp
volatile uint32_t t_avg;                      // average of 5 consecutive temperature readings (about 1 sec)
bool t_probe_present = false;                 // Temperature sensor is working
bool temp_probe_irq = false;                  // Tells us when a temperature conversion is complete
float t_dCel, t_dFahr;                        // Celsius and Farenheit temperature values
bool purging_valve = false;                   // True only when we're purging the hot water.
static int32_t desired_flow = 0x0;            // Desired flow 0=low, 1=med, 2=high
static int32_t absolute_max_temp = 106;       // absolute max temp F. System safety protection setting
static int32_t user_max_temp = 105;           // User set max temp F. User preference setting
                                              // Default software limit is 115 F (46 C). This is programmable by the user
                                              //    interface. This value will be saved in flash. Hardware
                                              //    safety temperature circuitry set to 120 F.
bool initialization_done = true;              // When the valve is first installed, the installation code
                                              //    will set this to "true".  This way we know all the system
                                              //    variables are configured properly
bool volatile static install_mode = false;    // First time installation(& reset)mode.
bool volatile static manual_mode = false;     // Need to know when the user is controlling things.
bool volatile static automatic_mode = false;  // not manual mode
bool volatile static program_mode = false;    // Used to know when we are recording a user programed profile.
bool volatile static program_run_mode = false; // Used to know when we are running a pre-programmed profile.
bool volatile static reset_ps = false;        // If resetWG = true then set initializationDone and installMOde
                                              //   to false so we can reset the valve and perform
                                              //   installation agian.
int32_t system_voltage;
static volatile bool clockwise_heat = true;
static volatile bool heat = true;
//static char heat_or_cool[10] = "heat" ;        //  heat or cool
//static char heat;        //  heat or cool
//static char cool;        //  heat or cool
bool user_interface_connect = false;          // connectetd to wireless controller
bool startup_done = false;                    // Delay to give the Nordic SD to time to initialize. Uses housekeeping timer
static volatile uint8_t step_pin, stepper_A, stepper_B;  // temporary value of output pin STEP_A & STEP_B
static volatile int8_t num_of_steps = 0;
//static volatile bool adjusting_valve = false;
static volatile bool stepping_forward = true;
static volatile bool controller_update_good = false;
static volatile uint32_t user_last_preset = 1;
static volatile bool user_presets_saved = false;
static volatile int shower_timer = 0;
static volatile int water_flow_rate = 2;
static volatile bool initial_command_sent = false;
static volatile int stepper_state = 0;
static volatile int  test_counter = 1;
static volatile uint8_t present_temp;         // nonvilatile version current_temp used to control temperaure
static volatile bool gen_timer_done = false;
static volatile bool flash_operation_completed = false;
static volatile bool update_flash = false;     // Use this to tell us when system settings changed and we need to save them.
static volatile uint32_t user_last_temp = 101;
static bool m_memory_access_in_progress;       //**< Flag to keep track of ongoing operations on persistent memory. */
static bool flash_test = true;                        // test flash code
static volatile bool purge_valve_now = false;

struct blockstruct // block to write into flash.
{
  int32_t desired_temp;               // Last water temperature requested by user.
  int32_t desired_flow;               
  int32_t user_max_temp;
  int32_t absolute_max_temp;
  int32_t startup_done;
  int32_t install_mode;
  int32_t initialization_done;
  int32_t clockwise_heat;
  int32_t shower_timer;
  int32_t temp_scale;
  int32_t user_last_preset;
  int32_t user_last_temp;
  int32_t flash_check;                // "0x55AA55AA" Keep at end of flash block. To verify flash is programmed.
}__ALIGN(4)write_flash_block, __ALIGN(4)read_flash_block;


/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

BLE_NUS_C_DEF(m_ble_nus_c);      /**< BLE Nordic UART Service (NUS) client instance. */
NRF_BLE_GATT_DEF(m_gatt);        /**< GATT module instance. */
BLE_DB_DISCOVERY_DEF(m_db_disc); /**< Database discovery module instance. */
NRF_BLE_SCAN_DEF(m_scan);        /**< Scanning Module instance. */

static uint16_t m_ble_nus_max_data_len = BLE_GATT_ATT_MTU_DEFAULT - OPCODE_LENGTH - HANDLE_LENGTH; /**< Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module. */

/**@brief NUS UUID. */
static ble_uuid_t const m_nus_uuid =
    {
        .uuid = BLE_UUID_NUS_SERVICE,
        .type = NUS_SERVICE_UUID_TYPE};

static char const m_target_periph_name[] = "WLC-01"; /**< Name of the device we try to connect to. This name is searched in the scan report data*/

/**@brief Function for handling asserts in the SoftDevice.
 *
 * @details This function is called in case of an assert in the SoftDevice.
 *
 * @warning This handler is only an example and is not meant for the final product. You need to analyze
 *          how your product is supposed to react in case of assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num     Line number of the failing assert call.
 * @param[in] p_file_name  File name of the failing assert call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t *p_file_name) {
  app_error_handler(0xDEADBEEF, line_num, p_file_name);
}

/**@brief Function for starting scanning. */
static void scan_start(void) {
  ret_code_t ret;

  ret = nrf_ble_scan_start(&m_scan);
  APP_ERROR_CHECK(ret);

  ret = bsp_indication_set(BSP_INDICATE_SCANNING);
  APP_ERROR_CHECK(ret);
}

/**@brief Function for handling Scanning Module events.
 */
static void scan_evt_handler(scan_evt_t const *p_scan_evt) {
  ret_code_t err_code;

  switch (p_scan_evt->scan_evt_id) {
  case NRF_BLE_SCAN_EVT_CONNECTING_ERROR: {
    err_code = p_scan_evt->params.connecting_err.err_code;
    APP_ERROR_CHECK(err_code);
  } break;

  case NRF_BLE_SCAN_EVT_CONNECTED: {
    ble_gap_evt_connected_t const *p_connected =
        p_scan_evt->params.connected.p_connected;
    // Scan is automatically stopped by the connection.
    NRF_LOG_INFO("Connecting to target %02x%02x%02x%02x%02x%02x",
        p_connected->peer_addr.addr[0],
        p_connected->peer_addr.addr[1],
        p_connected->peer_addr.addr[2],
        p_connected->peer_addr.addr[3],
        p_connected->peer_addr.addr[4],
        p_connected->peer_addr.addr[5]);
  } break;

  case NRF_BLE_SCAN_EVT_SCAN_TIMEOUT: {
    NRF_LOG_INFO("Scan timed out.");
    scan_start();
  } break;

  default:
    break;
  }
}

/**@brief Function for initializing the scanning and setting the filters.
 */
static void scan_init(void) {
  ret_code_t err_code;
  nrf_ble_scan_init_t init_scan;

  memset(&init_scan, 0, sizeof(init_scan));

  init_scan.connect_if_match = true;
  init_scan.conn_cfg_tag = APP_BLE_CONN_CFG_TAG;

  err_code = nrf_ble_scan_init(&m_scan, &init_scan, scan_evt_handler);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_ble_scan_filter_set(&m_scan, SCAN_NAME_FILTER, &m_target_periph_name);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_ble_scan_filters_enable(&m_scan, NRF_BLE_SCAN_NAME_FILTER, false);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling database discovery events.
 *
 * @details This function is a callback function to handle events from the database discovery module.
 *          Depending on the UUIDs that are discovered, this function forwards the events
 *          to their respective services.
 *
 * @param[in] p_event  Pointer to the database discovery event.
 */
static void db_disc_handler(ble_db_discovery_evt_t *p_evt) {
  ble_nus_c_on_db_disc_evt(&m_ble_nus_c, p_evt);
}

/**@brief Function for handling characters received by the Nordic UART Service (NUS).
 *
 * @details This function takes a list of characters of length data_len and prints the characters out on UART.
 *          If @ref ECHOBACK_BLE_UART_DATA is set, the data is sent back to sender.
 */
static void ble_nus_chars_received_uart_print(uint8_t *p_data, uint16_t data_len) {
  ret_code_t ret_val;

//  NRF_LOG_DEBUG("Receiving data in received_uart_print.");
  NRF_LOG_HEXDUMP_DEBUG(p_data, data_len);

  for (uint32_t i = 0; i < data_len; i++) {
    do {
      ret_val = app_uart_put(p_data[i]);
      if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY)) {
        NRF_LOG_ERROR("app_uart_put failed for index 0x%04x.", i);
        APP_ERROR_CHECK(ret_val);
      }
    } while (ret_val == NRF_ERROR_BUSY);
  }
  if (p_data[data_len - 1] == '\r') {
    while (app_uart_put('\n') == NRF_ERROR_BUSY)
      ;
  }
  if (ECHOBACK_BLE_UART_DATA) {
    // Send data back to the peripheral.
    do {
      ret_val = ble_nus_c_string_send(&m_ble_nus_c, p_data, data_len);
      if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY)) {
        NRF_LOG_ERROR("Failed sending NUS message. Error 0x%x. ", ret_val);
        APP_ERROR_CHECK(ret_val);
      }
    } while (ret_val == NRF_ERROR_BUSY);
  }
}

/**@brief   Function for handling app_uart events.
 *
 * @details This function receives a single character from the app_uart module and appends it to
 *          a string. The string is sent over BLE when the last character received is a
 *          'new line' '\n' (hex 0x0A) or if the string reaches the maximum data length.
 */
void uart_event_handle(app_uart_evt_t *p_event) {
  static uint8_t data_array[BLE_NUS_MAX_DATA_LEN];
  static uint16_t index = 0;
  uint32_t ret_val;

  switch (p_event->evt_type) {
  /**@snippet [Handling data from UART] */
  case APP_UART_DATA_READY:
    UNUSED_VARIABLE(app_uart_get(&data_array[index]));
    index++;

    if ((data_array[index - 1] == '\n') || (index >= (m_ble_nus_max_data_len))) {
      NRF_LOG_DEBUG("Ready to send data over BLE NUS");
      NRF_LOG_HEXDUMP_DEBUG(data_array, index);

      do {
        ret_val = ble_nus_c_string_send(&m_ble_nus_c, data_array, index);
        if ((ret_val != NRF_ERROR_INVALID_STATE) && (ret_val != NRF_ERROR_RESOURCES)) {
          APP_ERROR_CHECK(ret_val);
        }
      } while (ret_val == NRF_ERROR_RESOURCES);

      index = 0;
    }
    break;

  /**@snippet [Handling data from UART] */
  case APP_UART_COMMUNICATION_ERROR:
    NRF_LOG_ERROR("Communication error occurred while handling UART.");
    APP_ERROR_HANDLER(p_event->data.error_communication);
    break;

  case APP_UART_FIFO_ERROR:
    NRF_LOG_ERROR("Error occurred in FIFO module used by UART.");
    APP_ERROR_HANDLER(p_event->data.error_code);
    break;

  default:
    break;
  }
}
/***************************************************************************
**    Perfect Shower function for handling controller messages
****************************************************************************/

unsigned char sdata[20];
uint32_t temp_scale_C_F;

void ble_handle_message(char *pdata, int len) {
  //  unsigned char sdata[20];
  int i, ndata;

  // We Received BLE Communications from Wireless Controller
  //  Validate good data.

  if (len <= 5 ||
      pdata[0] != 0xfd || pdata[1] != 0xfc ||
      pdata[2] != len - 4) {
    NRF_LOG_INFO("Message reject point 341.");
    goto reject;
  }
  //  Validate check sum

  uint32_t cs = 0;
  for (i = 0; i < (3 + pdata[2]); i++)
    cs += pdata[i];
  cs = 0xff & ((~cs) + 1);
  if (cs != pdata[i]) {
    NRF_LOG_INFO("Message reject checksum, i %d, sb %x, was %x.", i, cs, pdata[i]);
    goto reject;
  }

  user_interface_connect = true;

  //  Analyze message per EPOB0178:

  // Reset Valve To Factory Settings Command (COMMAND #1)
  if (pdata[3] == 0x01 && pdata[2] == 1) {
    NRF_LOG_INFO("Remove valve data cmd.");
    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x81, 0x01, 0x00, 0x00, 0x00, 0x00);
    ndata = 5;
    sendreply(ndata);
  }

  //  Set Bluetooth Mode Command (COMMAND #2)
  else if (pdata[3] == 0x02 && pdata[2] == 1) {
    NRF_LOG_INFO("Set BT work mode cmd.");
    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x82, 0x01, 0x00, 0x00, 0x00, 0x00);
    ndata = 5;
    sendreply(ndata);
  }

  // Request Water Temperature From Valve Command (COMMAND #3)
  // The wireless controller sets the user's temperature scale that the valve uses.
  else if (pdata[3] == 0x03 && pdata[2] == 2) {
    NRF_LOG_INFO("Water temp query, scale %d.", pdata[4]);

    temp_scale_C_F = pdata[4]; // set the temp_scale, 0 (false)=F, 1 (true) = C
                           // NOTE if temperature scale changes, we need to purge temperature buffer and disable water temperature
                           // changes for at least one second.
    if (temp_scale_C_F != temp_scale)     // Change the temperature scale for the system.
    {
      temp_scale = temp_scale_C_F;
      write_flash_block.temp_scale = temp_scale;    // put new value into the flash_write structure
      update_flash = true;                          // Indicate that flash needs to be updated
      convert_temperature_values();
    }
//    uint8_t water_temp = (t_avg & 0x000000ff);     // Convert temp to 8 bits
//    uint8_t scale = (temp_scale & 0x000000ff);

    NRF_LOG_DEBUG("Sending Temp: %d \r\n", t_avg);
    SETSDATA(0xfd, 0xfc, 0x03, 0x83, temp_scale, t_avg, 0x00, 0x00, 0x00); // reply with current water temperature
    ndata = 6;
    sendreply(ndata);
  }

  // Set User Desired Water Temperature Command (COMMAND #4)
  else if (pdata[3] == 0x04 && pdata[2] == 3) {
    NRF_LOG_INFO("Water temp cmd, scale %d, degs %d.", pdata[4], pdata[5]);
    temp_scale_C_F = pdata[4]; // set the temp_scale, 0 (false)=F, 1 (true) = C
                           // NOTE if temperature scale changes, we need to purge temperature buffer and disable water temperature
                           // changes for at least one second.

    if (temp_scale_C_F != temp_scale)     // Change the temperature scale for the system.
    {
      temp_scale = temp_scale_C_F;
      write_flash_block.temp_scale = temp_scale;
      update_flash = true;
      convert_temperature_values();
    }

    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x84, 0x01, 0x00, 0x00, 0x00, 0x00);
    ndata = 5;
    sendreply(ndata);
  }
  // Set Desired Water Flow Rate (COMMAND #5)
  else if (pdata[3] == 0x05 && pdata[2] == 2) {
    NRF_LOG_INFO("Water flow cmd, flow %d.", pdata[4]);
    user_requested_flow(pdata[4]);

    // set flow rate here;

    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x85, 0x01, 0x00, 0x00, 0x00, 0x00);
    ndata = 5;
    sendreply(ndata);
  }

/****************** RESPONSES FROM CONTROLLER ACKNOWLEDGING VALVE COMMANDS *****************/
  // Water Flow Acknowledge Response from Controller
  else if (pdata[3] == 0x081 && pdata[2] == 5) {
    NRF_LOG_INFO("Water Flow Acknowledge Response, flow %d, scale %d, temp %d, timer %d.", pdata[4], pdata[5], pdata[6], pdata[7]);
    water_flow_rate = pdata[4];   // 0-low, 1-medium, 2-high
    temp_scale = pdata[5];        // set the temp_scale, 0 (false)=F, 1 (true) = C
//    desired_temp = pdata[6];
    shower_timer = pdata[7];
    initial_command_sent = true;
  }

  // Received Current Valve Settings - Response for Controller update command 02
  else if (pdata[3] == 0x082 && pdata[2] == 2) {
    NRF_LOG_INFO("Update Controller Settinge Response, flow %d.", pdata[4]);
    if (pdata[4] == 0){
      controller_update_good = false;
    } else {
      controller_update_good = true;
    }
  }
 
  // Update User Presets - Response from Controller
  else if (pdata[3] == 0x083 && pdata[2] == 2) {
    NRF_LOG_INFO("Update User Presets Response, flow %d.", pdata[4]);
    if (pdata[4] == 0){
      user_presets_saved = false;
    } else {
      user_presets_saved = true;
    }
  }

  else {
    NRF_LOG_INFO("Other command received, %x, %x, %x.", pdata[4], pdata[5], pdata[6]);
    //  Not replying. Ignore. Do nothing.
//    goto reject;
  }

  if (!initial_command_sent){
    send_to_controller("water_flowing");
  }

reject:
  return;
}

void send_to_controller(char *command, int data){
  int i, ndata;

  if (command == "water_flowing"){     //  Send Water Flowing
    SETSDATA(0xfd, 0xfc, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00);
    ndata = 4;
    sendreply(ndata);
  }

  else if (command == "update_controller_settings"){
    SETSDATA(0xfd, 0xfc, 0x05, 0x02, 0x02, 0x00, t_avg, 0x00, 0x00);
    ndata = 8;
    sendreply(ndata);
  }

  else if (command == "update_controller_presets"){
    SETSDATA(0xfd, 0xfc, 0x06, 0x03, 0x00, t_avg, (t_avg+20), (t_avg+5), (t_avg+15));
    ndata = 9;
    sendreply(ndata);
  }
}

void sendreply(int len_data) {
//  unsigned char sdata[20];
  int i;
  uint32_t cs;

  //  Compute checksum.
  cs = 0;
  for (i = 0; i < len_data; i++)
    cs += sdata[i];
  cs = 0xff & ((~cs) + 1);
  sdata[len_data++] = cs;

  //  Send reply.

  uint32_t ret_val = 21; // arbitrary marker number.
  do {
    ret_val = ble_nus_c_string_send(&m_ble_nus_c, sdata, len_data);
    if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY)) {
      NRF_LOG_ERROR("Failed sending NUS message. Error 0x%x. ", ret_val);
      APP_ERROR_CHECK(ret_val);
    }
  } while (ret_val == NRF_ERROR_BUSY);
  return;
}

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

/**@brief Callback handling Nordic UART Service (NUS) client events.
 *
 * @details This function is called to notify the application of NUS client events.
 *
 * @param[in]   p_ble_nus_c   NUS client handle. This identifies the NUS client.
 * @param[in]   p_ble_nus_evt Pointer to the NUS client event.
 */

/**@snippet [Handling events from the ble_nus_c module] */
static void ble_nus_c_evt_handler(ble_nus_c_t *p_ble_nus_c, ble_nus_c_evt_t const *p_ble_nus_evt) {
  ret_code_t err_code;

  switch (p_ble_nus_evt->evt_type) {
  case BLE_NUS_C_EVT_DISCOVERY_COMPLETE:
    NRF_LOG_INFO("Discovery complete.");
    err_code = ble_nus_c_handles_assign(p_ble_nus_c, p_ble_nus_evt->conn_handle, &p_ble_nus_evt->handles);
    APP_ERROR_CHECK(err_code);

    err_code = ble_nus_c_tx_notif_enable(p_ble_nus_c);
    APP_ERROR_CHECK(err_code);
    NRF_LOG_INFO("Connected to device with Nordic UART Service.");
    break;

  case BLE_NUS_C_EVT_NUS_TX_EVT:
    ble_nus_chars_received_uart_print(p_ble_nus_evt->p_data, p_ble_nus_evt->data_len);

    /******************************* DFH *************************/
    /******************************* DFH *************************/
    /******************************* DFH *************************/

    NRF_LOG_INFO("NUS TX EVT occurred.");

    ble_handle_message(p_ble_nus_evt->p_data, p_ble_nus_evt->data_len);

    /******************************* DFH *************************/
    /******************************* DFH *************************/
    /******************************* DFH *************************/

    break;

  case BLE_NUS_C_EVT_DISCONNECTED:
    NRF_LOG_INFO("Disconnected.");
    scan_start();
    break;
  }
}
/**@snippet [Handling events from the ble_nus_c module] */

/**
 * @brief Function for handling shutdown events.
 *
 * @param[in]   event       Shutdown type.
 */
static bool shutdown_handler(nrf_pwr_mgmt_evt_t event) {
  ret_code_t err_code;

  err_code = bsp_indication_set(BSP_INDICATE_IDLE);
  APP_ERROR_CHECK(err_code);

  switch (event) {
  case NRF_PWR_MGMT_EVT_PREPARE_WAKEUP:
    // Prepare wakeup buttons.
    err_code = bsp_btn_ble_sleep_mode_prepare();
    APP_ERROR_CHECK(err_code);
    break;

  default:
    break;
  }

  return true;
}

NRF_PWR_MGMT_HANDLER_REGISTER(shutdown_handler, APP_SHUTDOWN_HANDLER_PRIORITY);

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const *p_ble_evt, void *p_context) {
  ret_code_t err_code;
  ble_gap_evt_t const *p_gap_evt = &p_ble_evt->evt.gap_evt;

  switch (p_ble_evt->header.evt_id) {
  case BLE_GAP_EVT_CONNECTED:
    err_code = ble_nus_c_handles_assign(&m_ble_nus_c, p_ble_evt->evt.gap_evt.conn_handle, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_indication_set(BSP_INDICATE_CONNECTED);
    APP_ERROR_CHECK(err_code);

    // start discovery of services. The NUS Client waits for a discovery result
    err_code = ble_db_discovery_start(&m_db_disc, p_ble_evt->evt.gap_evt.conn_handle);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_DISCONNECTED:

    NRF_LOG_INFO("Disconnected. conn_handle: 0x%x, reason: 0x%x",
        p_gap_evt->conn_handle,
        p_gap_evt->params.disconnected.reason);
    break;

  case BLE_GAP_EVT_TIMEOUT:
    if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN) {
      NRF_LOG_INFO("Connection Request timed out.");
    }
    break;

  case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
    // Pairing not supported.
    err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
    // Accepting parameters requested by peer.
    err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
        &p_gap_evt->params.conn_param_update_request.conn_params);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_PHY_UPDATE_REQUEST: {
    NRF_LOG_DEBUG("PHY update request.");
    ble_gap_phys_t const phys =
        {
            .rx_phys = BLE_GAP_PHY_AUTO,
            .tx_phys = BLE_GAP_PHY_AUTO,
        };
    err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
    APP_ERROR_CHECK(err_code);
  } break;

  case BLE_GATTC_EVT_TIMEOUT:
    // Disconnect on GATT Client timeout event.
    NRF_LOG_DEBUG("GATT Client Timeout.");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTS_EVT_TIMEOUT:
    // Disconnect on GATT Server timeout event.
    NRF_LOG_DEBUG("GATT Server Timeout.");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  default:
    break;
  }
}

/*************************** Added by Eric *****************************/
  // flash storage handler
static void soc_evt_handler(uint32_t evt_id, void * p_context)
{
    switch (evt_id)
    {
        case NRF_EVT_FLASH_OPERATION_SUCCESS:
        // fall through 
          NRF_LOG_INFO("Flash write successful. \n");
        case NRF_EVT_FLASH_OPERATION_ERROR:

            if (flash_operation_completed)
            {
                flash_operation_completed = false;
            }
            break;

        default:
            // No implementation needed.
              flash_operation_completed = true;
            break;
    }
}
/* ******************************** End *********************************/

static void ble_stack_init(void) {
  ret_code_t err_code;
  uint32_t   ram_start = 0;

  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);

  // Configure the BLE stack using the default settings.
  // Fetch the start address of the application RAM.
//  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
  APP_ERROR_CHECK(err_code);

  // Enable BLE stack.
  err_code = nrf_sdh_ble_enable(&ram_start);
  APP_ERROR_CHECK(err_code);

 /*************************** Added by Eric ***********************/
 // Enable SOC.
//  err_code = nrf_sdh_soc_enable();
//  err_code = NRF_SDH_SOC_ENABLE();
//  APP_ERROR_CHECK(err_code);
/* ******************************** End *********************************/

  // Register a handler for BLE events.
  NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);

/*************************** Added by Eric ***********************/
  NRF_SDH_SOC_OBSERVER(m_soc_observer, APP_SOC_OBSERVER_PRIO, soc_evt_handler, NULL);
/*
  // Register with the SoftDevice handler module for BLE events.
  err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
  APP_ERROR_CHECK(err_code);

    /* Request Softdevice to turn on Crystal oscillator to test if SoC events are
       forwarded to the application. _SOC_ observers should get notified of the
       NRF_EVT_HFCLKSTARTED event after this call. */
    err_code = sd_clock_hfclk_request();
    APP_ERROR_CHECK(err_code);
/* ******************************** End *********************************/
}

/**@brief Function for handling events from the GATT library. */
void gatt_evt_handler(nrf_ble_gatt_t *p_gatt, nrf_ble_gatt_evt_t const *p_evt) {
  if (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED) {
    NRF_LOG_INFO("ATT MTU exchange completed.");

    m_ble_nus_max_data_len = p_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
    NRF_LOG_INFO("Ble NUS max data length set to 0x%X(%d)", m_ble_nus_max_data_len, m_ble_nus_max_data_len);
  }
}

/**@brief Function for initializing the GATT library. */
void gatt_init(void) {
  ret_code_t err_code;

  err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_ble_gatt_att_mtu_central_set(&m_gatt, NRF_SDH_BLE_GATT_MAX_MTU_SIZE);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling events from the BSP module.
 *
 * @param[in] event  Event generated by button press.
 */
void bsp_event_handler(bsp_event_t event) {
  ret_code_t err_code;

  switch (event) {
  case BSP_EVENT_SLEEP:
    nrf_pwr_mgmt_shutdown(NRF_PWR_MGMT_SHUTDOWN_GOTO_SYSOFF);
    break;

  case BSP_EVENT_DISCONNECT:
    err_code = sd_ble_gap_disconnect(m_ble_nus_c.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    if (err_code != NRF_ERROR_INVALID_STATE) {
      APP_ERROR_CHECK(err_code);
    }
    break;

  default:
    break;
  }
}

/**@brief Function for initializing the UART. */
static void uart_init(void) {
  ret_code_t err_code;

  app_uart_comm_params_t const comm_params =
      {
          .rx_pin_no = RX_PIN_NUMBER,
          .tx_pin_no = TX_PIN_NUMBER,
          .rts_pin_no = RTS_PIN_NUMBER,
          .cts_pin_no = CTS_PIN_NUMBER,
          .flow_control = APP_UART_FLOW_CONTROL_DISABLED,
          .use_parity = false,
          .baud_rate = UART_BAUDRATE_BAUDRATE_Baud115200};

  APP_UART_FIFO_INIT(&comm_params,
      UART_RX_BUF_SIZE,
      UART_TX_BUF_SIZE,
      uart_event_handle,
      APP_IRQ_PRIORITY_LOWEST,
      err_code);

  APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the Nordic UART Service (NUS) client. */
static void nus_c_init(void) {
  ret_code_t err_code;
  ble_nus_c_init_t init;

  init.evt_handler = ble_nus_c_evt_handler;

  err_code = ble_nus_c_init(&m_ble_nus_c, &init);
  APP_ERROR_CHECK(err_code);
}

/* Function for initializing timer, pins, buttons and leds. */
static void pins_buttons_leds_init(void) {
    ret_code_t err_code;
    bsp_event_t startup_event;

    err_code = bsp_init(BSP_INIT_LEDS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);

  // ************************************************************************
  // ***************** Pin Configuration - System I/Os **********************
  // ************************************************************************
  // NOTE: D_TEMP input, AKA ONEWIRE is is bidirectional and is controlled in temperature routine
    nrf_gpio_cfg_output(ONEWIRE_PIN);  
    nrf_gpio_pin_set(ONEWIRE_PIN);

    // Set STEP_A and STEP_B pins as standard outputs and set the outputs low
//    nrf_gpio_cfg_output(STEP_A); // Set pins as outputs GPIO_PIN_CNF_DRIVE_H0H1

      NRF_GPIO->PIN_CNF[STEP_A] = 0 | (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos) 
                                     | (GPIO_PIN_CNF_DRIVE_H0H1 << GPIO_PIN_CNF_DRIVE_Pos) 
                                     | (GPIO_PIN_CNF_PULL_Disabled << GPIO_PIN_CNF_PULL_Pos) 
                                     | (GPIO_PIN_CNF_INPUT_Connect << GPIO_PIN_CNF_INPUT_Pos) 
                                     | (GPIO_PIN_CNF_DIR_Output << GPIO_PIN_CNF_DIR_Pos);

//    nrf_gpio_cfg_output(STEP_B);
      NRF_GPIO->PIN_CNF[STEP_B] = 0 | (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos) 
                                     | (GPIO_PIN_CNF_DRIVE_H0H1 << GPIO_PIN_CNF_DRIVE_Pos) 
                                     | (GPIO_PIN_CNF_PULL_Disabled << GPIO_PIN_CNF_PULL_Pos) 
                                     | (GPIO_PIN_CNF_INPUT_Connect << GPIO_PIN_CNF_INPUT_Pos) 
                                     | (GPIO_PIN_CNF_DIR_Output << GPIO_PIN_CNF_DIR_Pos);

    nrf_gpio_pin_clear(STEP_A); // Set pins low.
    nrf_gpio_pin_clear(STEP_B);

    // Set up buzzer pin as standard output.
      NRF_GPIO->PIN_CNF[AUDIO_OUT] = 0 | (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos) 
                                     | (GPIO_PIN_CNF_DRIVE_S0S1 << GPIO_PIN_CNF_DRIVE_Pos) 
                                     | (GPIO_PIN_CNF_PULL_Disabled << GPIO_PIN_CNF_PULL_Pos) 
                                     | (GPIO_PIN_CNF_INPUT_Connect << GPIO_PIN_CNF_INPUT_Pos) 
                                     | (GPIO_PIN_CNF_DIR_Output << GPIO_PIN_CNF_DIR_Pos);
    nrf_gpio_cfg_output(AUDIO_OUT);
    nrf_gpio_pin_clear(AUDIO_OUT);         // Set pin low.
//    nrf_gpio_pin_set(AUDIO_OUT);           // Set pin high.

    // Set up LEDs as standard outputs.
    nrf_gpio_cfg_output(LED_1);
    nrf_gpio_pin_clear(LED_1); // Set pin low. Turns on LED.
    nrf_gpio_cfg_output(LED_2);
    nrf_gpio_pin_clear(LED_2); // Set pin low. Turns on LED.

    // Set up a_temp_out as standard output
    nrf_gpio_cfg_output(A_TEMP_OUT);
    nrf_gpio_pin_clear(A_TEMP_OUT); // Set pin low.

    // Set up Sys_V as standard input w/ no pullup.
    nrf_gpio_cfg_input(SYS_V, NRF_GPIO_PIN_NOPULL);

    // Set up A_temp_in as standard input w/ no pullup.
    nrf_gpio_cfg_input(A_TEMP_IN, NRF_GPIO_PIN_NOPULL);

    // Set up trip_test as a standard output with pull down;
    NRF_GPIO->PIN_CNF[TRIP_TEST] = (NRF_GPIO_PIN_NOSENSE) | (NRF_GPIO_PIN_S0S1) | (NRF_GPIO_PIN_PULLDOWN) | (NRF_GPIO_PIN_INPUT_DISCONNECT) | (NRF_GPIO_PIN_DIR_OUTPUT);

    // Set up temp_override pin as open-collector high drive output
    NRF_GPIO->PIN_CNF[TEMP_OVERRIDE] = (NRF_GPIO_PIN_NOSENSE) | (NRF_GPIO_PIN_H0D1) | (NRF_GPIO_PIN_NOPULL) | (NRF_GPIO_PIN_INPUT_DISCONNECT) | (NRF_GPIO_PIN_DIR_OUTPUT);
    nrf_gpio_pin_set(TEMP_OVERRIDE); // Set pin low to disable stepper.

    // Set up manual override and manual reset pins as sensing inputs w/pullups
    nrf_gpio_cfg_sense_input(MAN_OVERRIDE, NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_HIGH); // SENSE_HIGH OR LOW?
    nrf_gpio_cfg_sense_input(MAN_RESET, NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_HIGH);    // SENSE_HIGH OR LOW?
    nrf_gpio_cfg_sense_input(VALVE_OFF, NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_HIGH);    // SENSE_HIGH OR LOW?
}
/******************** END OF PIN CONFIGURATIONS *************************/
//******************************************************************************************
/*************** Code for Interrupts (not working properly) - (NOT USED) *******************
//******************************************************************************************/
/*
// GPIOTE is used for D_TEMP_IRQ input to know when a temperature conversion is completed
void in_pin_handler(nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action) {
  get_water_temp();
}
/**
 * @brief Function for configuring: PIN_IN pin for input, 
 * and configures GPIOTE to give an interrupt on pin change. 

#define PIN_IN D_TEMP_IRQ

void gpio_init(void) {
  ret_code_t err_code;
  err_code = nrf_drv_gpiote_init();
  APP_ERROR_CHECK(err_code);

  nrf_drv_gpiote_in_config_t in_config = GPIOTE_CONFIG_IN_SENSE_LOTOHI(true);
  in_config.pull = NRF_GPIO_PIN_NOPULL;

  err_code = nrf_drv_gpiote_in_init(PIN_IN, &in_config, in_pin_handler);
  APP_ERROR_CHECK(err_code);

//  d_temp_irq_enable();
}
//  Functions to enable and disable interrupts on d_temp_irq
int d_temp_irq_enable(void) {
  nrf_drv_gpiote_in_event_enable(PIN_IN, true);
//  temp_probe_irq = true;
}

uint32_t d_temp_irq_disable(void) {
  nrf_drv_gpiote_in_event_disable(PIN_IN);
//  temp_probe_irq = false;
}
/*****************************************************************************************
                Set up timers for our general use
******************************************************************************************/
APP_TIMER_DEF(startup_timer);
static void startup_timer_handler(void * p_context)
{
  startup_done = true;          // Used to give the Nordic SD time to finish
}


APP_TIMER_DEF(read_temperature_timer);
static void read_temperature_timer_handler(void * p_context)
{
//  nrf_gpio_pin_toggle(LED_2); // Flash the led so we know something's happening
  get_water_temp();
}


APP_TIMER_DEF(stepper_delay_timer);
void stepper_delay_timer_handler(void *p_context) 
//void stepper_delay_timer_handler(nrf_timer_event_t event_type, void *p_context) 
{
  if ((install_mode) || (purge_valve_now))
  {
    rotate_valve();
  } else 
  {
    start_stepper_timer();
    adjust_temp(desired_temp);
  }
}


APP_TIMER_DEF(check_wireless_comm_timer);
static void check_wireless_comm_timer_handler(void * p_context)
{
  if (!user_interface_connect){               // keep checking until comms established
    start_check_wireless_comm_timer();
  }
}

static bool gen_timer_expire = false;

APP_TIMER_DEF(gen_timer);           // Use as a general timer - Multiple useages. 
static void gen_timer_handler(void * p_context)
{
  gen_timer_done = true;          // Used to give the Nordic SD time to finish
  // Multiple useages
  gen_timer_expire = true;

}

/**************************** SAADC Timer Section *************************/
#define SAMPLES_IN_BUFFER 5
volatile uint8_t state = 1;

static const nrf_drv_timer_t m_timer = NRF_DRV_TIMER_INSTANCE(1); // Eric - changed from 0 to 1 since 0 is used by SD. Tiimer 2 used by PWM
static nrf_saadc_value_t m_buffer_pool[2][SAMPLES_IN_BUFFER];
static nrf_ppi_channel_t m_ppi_channel;
static uint32_t m_adc_evt_counter;

//  APP_TIMER_DEF(saadc_timer);   // Eric - added this line. Code doesn't seem matter.
void saadc_timer_handler(nrf_timer_event_t event_type, void *p_context) {
  saadc_callback();
  printf("saadc timer timed out");
}

static void app_timers_init(void) 
{
  // Create timers
  ret_code_t err_code;
  err_code = app_timer_create(&startup_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              startup_timer_handler);

  err_code = app_timer_create(&read_temperature_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              read_temperature_timer_handler);

  err_code = app_timer_create(&stepper_delay_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              stepper_delay_timer_handler);

  err_code = app_timer_create(&check_wireless_comm_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              check_wireless_comm_timer_handler);

  err_code = app_timer_create(&gen_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              gen_timer_handler);
}

// Start housekeeping timer - Delay for BLE to finish before starting water temperature control
void start_startup_timer()
{
  uint32_t err_code;
  err_code = app_timer_start(startup_timer, APP_TIMER_TICKS(STARTUP_TIME), NULL);
  APP_ERROR_CHECK(err_code);
}

// Start the Temperature Conversion timer
void start_temperature_conversion_timer(convert_time)
{
  uint32_t time_in_ticks;
//  time_in_ticks = nrf_drv_timer_ms_to_ticks(&read_temperature_timer, convert_time);
  time_in_ticks = convert_time;

  uint32_t err_code;
  err_code = app_timer_start(read_temperature_timer, APP_TIMER_TICKS(time_in_ticks), NULL);
  APP_ERROR_CHECK(err_code);
}
// Start Stepper timer
void start_stepper_timer(void) 
{
  ret_code_t err_code;
  err_code = app_timer_start(stepper_delay_timer, APP_TIMER_TICKS(stepper_motor_rate), NULL); // mS between steps
  APP_ERROR_CHECK(err_code);
}

// Start the Communications Check timer. Timer checks about every half second for comms
void start_check_wireless_comm_timer()
{
  uint32_t err_code;
  err_code = app_timer_start(check_wireless_comm_timer, APP_TIMER_TICKS(500), NULL);
  APP_ERROR_CHECK(err_code);
}

// Start the flash timer. Timer checks about every 10mS for the flash to be done.
void start_general_timer()
{
  uint32_t err_code;
  err_code = app_timer_start(gen_timer, APP_TIMER_TICKS(100), NULL);
  APP_ERROR_CHECK(err_code);
}

/******************** END OF Timers CONFIGURATIONS *************************/

/**@brief Function for initializing the nrf log module. */
static void log_init(void) {
  ret_code_t err_code = NRF_LOG_INIT(NULL);
  APP_ERROR_CHECK(err_code);

  NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/**@brief Function for initializing power management.
 */
static void power_management_init(void) {
  ret_code_t err_code;
  err_code = nrf_pwr_mgmt_init();
  APP_ERROR_CHECK(err_code);
}

/** @brief Function for initializing the database discovery module. */
static void db_discovery_init(void) {
  ret_code_t err_code = ble_db_discovery_init(db_disc_handler);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling the idle state (main loop).
 *
 * @details Handles any pending log operations, then sleeps until the next event occurs.
 */
static void idle_state_handle(void) {
  if (NRF_LOG_PROCESS() == false) {
    nrf_pwr_mgmt_run();
  }
}

int main(void) {
  // Initialize.
  log_init();
  app_timer_init();              // Initialize timer lib
  app_timers_init();             // Setup the system timers
  uart_init();
  pins_buttons_leds_init();      
//  gpio_init();                 // Was used for Interrupts. NO LONGER USED
//  flash_init();                  
  pwm_init();                    
  saadc_init();                  
  saadc_sampling_event_init();   
  db_discovery_init();
  power_management_init();
  ble_stack_init();

// NOTE: OK to disable gatt_init() for testing (BLE won't work)

//  gatt_init();      

// NOTE: OK to disable nus_c_init() for testing (BLE won't work)
  nus_c_init();

// NOTE: OK to disable scan_init() for testing 
//  scan_init();

  // Start execution.
//  printf("BLE UART central example started.\r\n");
//  NRF_LOG_INFO("BLE UART central example started.");

// NOTE: OK to disable scan_start() for testing 
//  scan_start();


  start_startup_timer();
  flash_init();
  perfect_shower_init();         // Eric

  // Enter main loop.
  for (;;) {
    idle_state_handle();
  }
}
// *************************************************************************************
/* **************** End of Nordic ble_app_uart_c Example Code **************************
// *************************************************************************************/

// *************************************************************************************
/* **************** Start of Perfect Shower Main Routines **************************
// *************************************************************************************/
// *************** We don't control flow rate yet so we'll ignore for now **************
void user_requested_flow(int flow) {
  write_flash_block.desired_flow = flow;
  //  We'll use this function to turn on the valve
}
//
/*
void write_flash_block() {
  // bunch of NRF stuff.
}

void read_flash_block() {
  // bunch of NRF stuff.
  temp_scale = false; // C or F. True = C, False = F. This is set by the user via the display
}
*/
// *************************************************************************************
//                        End - Perfect Shower Main Code
// *************************************************************************************
// *************************************************************************************
//       Nordic DS18B20 Temperature sensor interface (FINISHED) (Eric's version)
// *************************************************************************************

enum {
  BSPACM_ONEWIRE_CMD_READ_ROM = 0x33,          // Read 64-bit ROM code without using search procedure
  BSPACM_ONEWIRE_CMD_SKIP_ROM = 0xCC,          // Skip ROM sends the following command to all bus devices
  BSPACM_ONEWIRE_CMD_READ_POWER_SUPPLY = 0xB4, // Tw]ƬOHͨѹq٬O~ѹq
  BSPACM_ONEWIRE_CMD_RECALL_EE = 0xB8,         // Store data from EEPROM into RAM
  BSPACM_ONEWIRE_CMD_READ_SCRATCHPAD = 0xBE,   // Read the RAM area.
  BSPACM_ONEWIRE_CMD_CONVERT_T = 0x44,         // Initiate a temperature conversion.
  BSPACM_ONEWIRE_CMD_WRITE_SCRATCHPAD = 0x4E,  // Write to scratchpad memory command
};

enum {
  OWT_RSTL_us = 480,                          // Minimum time to hold bus low to ensure reset
  OWT_RSTH_us = 480,                          // Time to wait for presence detection after reset to quiesce
  OWT_PDHIGH_us = 60,                         // Delay before presence pulse is known to be valid (15us..60us)
  OWT_LOW0_us = 60,                           // Minimum time to hold bus low when writing a zero
  OWT_LOW1_us = 5,                            // Eric was 1   	// Minimum time to hold bus low when writing a one
  OWT_REC_us = 5,                             // Eric was 1	// Recovery delay between write/read transaction cycles
  OWT_INT_us = 5,                             // Eric was 1	// Time to hold bus low when initiating a read slot
  OWT_RDV_us = 25 - OWT_INT_us,               // Eric was 15	// Point at which read value should be sampled
  OWT_SLOT_us = 80,                           // Eric was 60	// Minimum duration of a read or write slot
  BSPACM_ONEWIRE_CMD_WRITE_Th_Reg = 0x34,     // 34 = two's complement of 52C - the high temperature threshold
  BSPACM_ONEWIRE_CMD_WRITE_Tl_Reg = 0x00,     // 00 = two's complement of 0C - the Low temperature threshold
  BSPACM_ONEWIRE_CMD_WRITE_CONFIG_Reg = 0x3F, // Set up config reg for 10 bit conversion.
  TEMP_CONVERT_TIME_9BIT = 90,                // Time in ms to complete temperature conversion.
  TEMP_CONVERT_TIME_10BIT = 190,              // s/b 190 - Time in ms to complete temperature conversion.
  TEMP_CONVERT_TIME_11BIT = 380,              // Time in ms to complete temperature conversion.
  TEMP_CONVERT_TIME_12BIT = 755,              // Time in ms to complete temperature conversion.
};

#define ONEWIRE_BIT (1ULL << ONEWIRE_PIN)
static void dirset(void) // Set's pin as output.
{
  NRF_GPIO->DIRSET = ONEWIRE_BIT;
}

void dirclr(void) // Set's pin as input.
{
  NRF_GPIO->PIN_CNF[ONEWIRE_PIN] = 0 | (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos) 
                                     | (GPIO_PIN_CNF_DRIVE_S0S1 << GPIO_PIN_CNF_DRIVE_Pos) 
                                     | (GPIO_PIN_CNF_PULL_Disabled << GPIO_PIN_CNF_PULL_Pos) 
                                     | (GPIO_PIN_CNF_INPUT_Connect << GPIO_PIN_CNF_INPUT_Pos) 
                                     | (GPIO_PIN_CNF_DIR_Input << GPIO_PIN_CNF_DIR_Pos);
}

/* ************************** Set up the system at power up ************************
************************************************************************************/

void perfect_shower_init(void) {

/*************** READ THE FLASH SETTINGS **************************
*
* call "read flash function" here to get system settings.
*
*******************************************************************/

/***************  First get the temperature sensor working  *********/
    t_probe_present = false;
    if (0 == iBSPACMonewireReset_ni()) {
      printf("ERR: No DS18B20 present on P0.%u\n", ONEWIRE_PIN);
      NRF_LOG_INFO("ERR: No DS18B20 present on P0.%u\n", ONEWIRE_PIN);
      beep(20, 1000, 250);
      goto exit_probe_init;
    } else {                                                           
      t_probe_present = true;                                          // Temp device responded.
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);         // Send the SKIP ROM command.
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_SCRATCHPAD); // Write command
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_Th_Reg);     // Set the high temperature threshold
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_Tl_Reg);     // Set the low temperature threshold
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_CONFIG_Reg); // Write 3F to configuration register for 10 bit res
    }
    beep(2, 330, 100);           // Beep so we know the temperature sensor is working.
      // restart the temperature conversion process
    perform_temp_conversion();
    start_temperature_conversion_timer(TEMP_CONVERT_TIME_10BIT);

// Set the stepper_state to state 0
    nrf_gpio_pin_clear(STEP_A);
    nrf_gpio_pin_clear(STEP_B);
    stepper_state = 0;

                          /*    // Wait for valve to communicate with wireless controller
                              start_check_wireless_comm_timer();
// disable for testing                              while (!user_interface_connect)
                              {
                                beep(1, 300, 500);
                              }
                          */
    if (install_mode || !initialization_done || reset_ps) {  // Initialize ps.
  /*************************************************************************
   *                  VALVE ROTATION DIRECTION
   *  This code figures out which direction of rotation makes the water
   *  hotter and colder.  The valve could be plumbed backwards and different
   *  manufacturers have different designs or the stepper motor could be
   *  installed in the WaterGeni upside down.
   *  NOTE: This code is only run during initial installation or after a
   *  reset.  The "bool clockwiseHeat = true/false" is stored in flash for 
   *  subsequent reference during power up so the valve knows which way to turn.
   *
   *  The valve is first powered up after installation, by manually pulling out 
   *  the handle and rotating it.  The valve will be in "Manual" mode because
   *  the handle is pulled out to engage the handle and water will be flowing.  
   *  If the handle is not pulled out the handle will not be engaged and the 
   *  valve will not turn on.  Pull out handle and turn approximately 50%.   
   *  The valve will reengage to automatic mode by itself.
   *************************************************************************/
    NRF_LOG_INFO("Perfect Shower Install Started.\n");

    // Check to see if flash has already been initialized.  When factory new, flash is blank
    get_flash_data();

recheck_flash:
    if ((read_flash_block.flash_check != 0x55AA55AA) || reset_ps)       // Flash IS BLANK.
// if (reset_ps)   // use reset_ps to force a flash write during dev.
    {
      if (!temp_scale)     // False (0) = Farenheit
        {
          //  Initialize Flash - Fill flash structure with valid F data to initialize the system
          desired_temp = 80;                // dec F
          desired_flow = 3;    
          user_max_temp = 102;              // dec F
          absolute_max_temp = 105;          // dec F
          startup_done = false;
          install_mode = false;              // Set to 1 for development. Clear for production.
          initialization_done = true;      // Clear to 0 for development. Set for production.
          clockwise_heat = true;            // Sets valve rotation direction.
          shower_timer = 0;                 // Set to 0 minutes. 0 = no timer, 1 to 60 min max.
          temp_scale = false;               // Set to F initially
          user_last_preset = 1;
          user_last_temp = 101;             // dec F
        }

      if (temp_scale)    // 1 = Celsius
        {
          //  Initialize Flash - Fill flash structure with valid C data to initialize the system
          desired_temp = 74;                // dec value of 37C shifted left 1 bit for decimal point
          desired_flow = 3;    
          user_max_temp = 79;              // dec for 39C shifted left 1 bit
          absolute_max_temp = 81;          // dec for 40.5C shifted left 1 bit
          startup_done = false;
          install_mode = false;              // Set to 1 for development. Clear for production.
          initialization_done = true;      // Clear to 0 for development. Set for production.
          clockwise_heat = true;            // Sets valve rotation direction.
          shower_timer = 0;                 // Set to 0 minutes. 0 = no timer, 1 to 60 min max.
          temp_scale = false;               // Set to F initially
          user_last_preset = 1;
          user_last_temp = 76;              // dec for 38C shifted left 1 bit
        }

      install_mode = false;
      initialization_done = true;
      update_flash = true;
      update_flash_info();
      goto recheck_flash;
    }

    // copy flash data into working system variables
    desired_temp = read_flash_block.desired_temp;       // Last water temperature requested by user.
    desired_flow = read_flash_block.desired_flow;    
    user_max_temp = read_flash_block.user_max_temp;
    absolute_max_temp = read_flash_block.absolute_max_temp;
    startup_done = read_flash_block.startup_done;
    install_mode = read_flash_block.install_mode;
    initialization_done = read_flash_block.initialization_done;
    clockwise_heat = read_flash_block.clockwise_heat;
    shower_timer = read_flash_block.shower_timer;
    temp_scale = read_flash_block.temp_scale;
    user_last_preset = read_flash_block.user_last_preset;
    user_last_temp = read_flash_block.user_last_temp;
//    read_flash_block.flash_check;       // For validity check only. No reason to copy to sys variable

    start_stepper_timer();                  // Starting the timer begins the temperature control routine

        // User turned on valve approx 50%.  Since the valve has a 95 degree rotation from 
        // off to full hot and there 200 steps per rotation on the stepper, are there are
        // 1.8 degrees per step.  That means that there are approximately 53 steps from
        // off to full hot.  50% is about 26 steps.
retry_valve_direction:
    stabilize_temperature();    
    int32_t curr_temp_1 = present_temp;   // Find out what the water temp is.
    heat=true;
    rotate_valve(10);             // Turn the valve towrds what we think is HOT.
    stabilize_temperature();    
    int32_t curr_temp_2 = present_temp;   // get a second temp reading to compare with first reading.
    heat=false;
    rotate_valve(20);             // Turn the valve towrds what we think is HOT.
    stabilize_temperature();    
    int32_t curr_temp_3 = present_temp;   // get a second temp reading to compare with first reading.

    if ((curr_temp_2 > curr_temp_1) && (curr_temp_3 < curr_temp_1))
    {
      clockwise_heat = true;
    }else if ((curr_temp_2 < curr_temp_1) && (curr_temp_3 > curr_temp_1))
    {
      clockwise_heat = false;
    }else
    {
      goto retry_valve_direction;
    }
   
//******************** We know what direction the valve works now **********************/

    heat=true;
    rotate_valve(60);             // turn it enough to make sure we are at full hot OR OFF if valve backwards.
    stabilize_temperature();            // wait for water temperature to stabilize.
    curr_temp_1 = present_temp;         // Measure the maximum water temperature of the HOT water
    NRF_LOG_INFO("Maximum hot water temperature is %d.\n", curr_temp_1);

//********************* Write Wireless Message Code *************************
// Send hot water temperature message wireless controller.


//********************* Locate the valve and handle positions *************************
// Need to locate the reed switch positions so we know when the valve is about to 
// turn off or when it is placed into manual mode.


   // Measure system voltage (Sys_V) pin. Will be approx 1.0V when valve ON and water flowing.


//  Save initialization settings to flash here- Call Flash Routine here 
    install_mode = false;
    initialization_done = true;
    update_flash = true;
    update_flash_info();

      // ********************************************************************
    } // ******************* END OF VALVE INSTALL MODE **********************
      // ********************************************************************

    get_flash_data();

/************* disable for dev purposes ***********************
verify_flash:
    if (read_flash_block.flash_check != 0x55AA55AA)       // Flash IS BLANK.
// if (reset_ps)   // use reset_ps to force a flash write during dev.
    {
      install_mode = false;
      initialization_done = true;
      update_flash = true;
      update_flash_info();
      perfect_shower_init();                     // restart valve
      goto exit_probe_init;
    }
*/

    // copy flash data into working system variables
    desired_temp = read_flash_block.desired_temp;       // Last water temperature requested by user.
    desired_flow = read_flash_block.desired_flow;    
    user_max_temp = read_flash_block.user_max_temp;
    absolute_max_temp = read_flash_block.absolute_max_temp;
    startup_done = read_flash_block.startup_done;
    install_mode = read_flash_block.install_mode;
    initialization_done = read_flash_block.initialization_done;
    clockwise_heat = read_flash_block.clockwise_heat;
    shower_timer = read_flash_block.shower_timer;
    temp_scale = read_flash_block.temp_scale;
    user_last_preset = read_flash_block.user_last_preset;
    user_last_temp = read_flash_block.user_last_temp;

//    start_stepper_timer();                  // Starting the timer begins the temperature control routine
    
    purge_valve_now = true;                 // set flag for stepper timer handler
    purge_valve();                          // Flush the hot water
//    stabilize_temperature();                // Wait for hot and cold water to stabilize at whatever temp. 

exit_probe_init:

  return 99; // Send some error code.

} // End of initialization
/**************************** End of initialization ****************************/

/***************  Purge the hot water supply  **********************/

void purge_valve(void)
{
    uint8_t purge_steps = 60;
    heat = true;
//    rotate_valve(purge_steps);    // turn valve all the way to full hot. Valve stops at full ON.

    do{
      rotate_valve(1);    // turn valve all the way to full hot. Valve stops at full ON.
      purge_steps --;                     // Count the steps.
    }while (purge_steps != 0);

    purge_valve_now = false;
    stabilize_temperature();                // Wait for hot and cold water to stabilize at whatever temp. 
    adjust_temp(desired_temp);
}

/**************  Wait for water temp to stabilize ******************/
void stabilize_temperature(void)
{
  int32_t count_to_3 = 0;
  int32_t count_to_five = 0;
  int32_t temp_1, temp_2;
    // wait for the water temperature to stabilize for three consecutive readings 1 sec apart
    do{
       do{
          temp_1 = t_avg;                   // use the average temperature for smoothing purposes
         do{
            start_general_timer();             
            while(!gen_timer_expire);                   
            gen_timer_expire = false;
            count_to_five ++;
         } while (count_to_five < 6);      // wait about half a second then take another reading
          gen_timer_expire = false;
          temp_2 = t_avg;
       } while (temp_2 != temp_1);
        count_to_3 ++;
    } while (count_to_3 < 4);  
}

int get_water_temp(void) {
//  ***** This routine is ONLY entered via the read_temperture_timer timeout.
//  ***** All other system temperature functions should use global variables such as: 
//  ***** desired_temp, user_temp, raw_temp, current_temp and t_ave, some of which are set herein.
//  ***** This routine runs constantly in the background at the speed of the Dallas sensor.

  int32_t raw_temp;                                      // raw DS18B20 temperature sensor data
  // Temperature ready - Read it and convert it.
  t_raw = read_temp_probe(&t_raw);
  raw_temp = t_raw;
  raw_temp = raw_temp & ~3;                              // ignore last two bits (for 10 bit precicion)
  t_dCel = (float)raw_temp / 16.0;
  t_dFahr = t_dCel * 1.8 + 32.0;

    if (temp_scale) {                                        // for Celsius...
     int intpart = (int)t_dCel;
     float decimal_part = t_dCel - intpart;                 // just the decimal part
      if ((decimal_part > 0.5) && (decimal_part < 0.9)) {
        current_temp = ((intpart << 1)| 0x0001);              // set bit 0 to a 1 for 0.5C
      } else {
        current_temp = (intpart << 1);
      }
    } else {                                                 // for farenheit... Round up
      int intpart = (int)t_dFahr;
      float decimal_part = t_dFahr - intpart;                // just the decimal part
        if ((decimal_part > 0.5) && (decimal_part < 0.9)) {
          current_temp = t_dFahr + 1;                          // round up
        } else {
          current_temp = t_dFahr;                              // We only send intergers for F
        }
    }
      // For some reason (might be noise from wires in dev system), the temperature probe 
      // sometimes returns whacked-out high readings. 
      // Check for bad temperature readings and limit the temperature to a max difference of 
      // 10 degrees so we don't screw up our avg temp calc or our system response time.
    if (temp_scale && (current_temp >= (absolute_max_temp << 1)) || (!temp_scale && (current_temp >= (absolute_max_temp))))
    {
        goto skip_bad_value;
    }
  present_temp = current_temp;                             // use present_temp for adjusting stepper motor 
  tdata[t_count] = current_temp;                           // place in storage and calculate new average  
  if (t_count > 4) {                                       // 5 readings is about a second
    t_count = 0;
  } else {
    t_count++;
  }
      // Average the 3 readings.  Three readings chosen for a fast response time.
  t_avg = ((tdata[0] + tdata[1] + tdata[2] + tdata[3] + tdata[4]) / 5);

// Debug code:
//  if ((t_avg < 68) || (t_avg > 98)){
//    NRF_LOG_DEBUG("Temp out of range: \r\n %d %d %d %d %d %d", t_avg, tdata[0], tdata[1], tdata[2], tdata[3], tdata[4]);
//    beep(5,75,50);
//  }

skip_bad_value:
      // Now, restart a new temperature conversion cycle
  perform_temp_conversion();
  start_temperature_conversion_timer(TEMP_CONVERT_TIME_10BIT);
      // Go and adjust temperature while waiting for another temperature measurement
//    if (user_interface_connect) adjust_temp(desired_temp);

//test_counter = 1;
//adjust_temp(desired_temp);

//  return t_avg;
}


void convert_temperature_values(void)
{
/*  Copy of storage structure for referencing temperature values that need to be updated
  int32_t desired_temp;               // Last water temperature requested by user.
  int32_t user_max_temp;
  int32_t absolute_max_temp;
  int32_t user_last_temp;
*/ 
    if (temp_scale)         // Convert current temperature values to C
    {
      // t_dCel = (t_dFahr-32)/1.8;
      tdata[0] = (tdata[0]-32)/1.8;
      tdata[1] = (tdata[1]-32)/1.8;
      tdata[2] = (tdata[2]-32)/1.8;
      tdata[3] = (tdata[3]-32)/1.8;
      tdata[4] = (tdata[4]-32)/1.8;
      t_avg = ((tdata[0] + tdata[1] + tdata[2] + tdata[3] + tdata[4]) / 5);

      // Now convert all system variables to C
      desired_temp = (desired_temp - 32)/1.8;
      user_max_temp = (user_max_temp - 32)/1.8;
      absolute_max_temp = (absolute_max_temp - 32)/1.8;
      user_last_temp = (user_last_temp - 32)/1.8;
    } 
    else                    // Convert current temperature values to F
    {
      // t_dFahr = t_dCel * 1.8 + 32.0;
      tdata[0] = (tdata[0] * 1.8) + 32;
      tdata[1] = (tdata[1] * 1.8) + 32;
      tdata[2] = (tdata[2] * 1.8) + 32;
      tdata[3] = (tdata[3] * 1.8) + 32;
      tdata[4] = (tdata[4] * 1.8) + 32;
      t_avg = ((tdata[0] + tdata[1] + tdata[2] + tdata[3] + tdata[4]) / 5);
    
      // Now convert all system variables to F
      desired_temp = (desired_temp * 1.8) + 32;
      user_max_temp = (user_max_temp * 1.8) + 32;
      absolute_max_temp = (absolute_max_temp * 1.8) + 32;
      user_last_temp = (user_last_temp * 1.8) + 32;
    }
    
    // Change all of the temperature data values in working memory and flash_write structure
    write_flash_block.desired_temp = desired_temp;    // put new value into the flash_write structure
    write_flash_block.user_max_temp = user_max_temp;    // put new value into the flash_write structure
    write_flash_block.absolute_max_temp = absolute_max_temp;    // put new value into the flash_write structure
    write_flash_block.user_last_temp = user_last_temp;    // put new value into the flash_write structure

    update_flash = true;
    update_flash_info();
}


//  ************* Dallas Temp Sensor Reset  ********************
int iBSPACMonewireReset_ni(void) {
  int present = 0;

  dirset();
  NRF_GPIO->OUTSET = ONEWIRE_BIT; // Set pin high
  nrf_delay_us(480);              // Start by holding bus high for 480uS

  // Hold bus low for OWT_RESET_us
  NRF_GPIO->OUTCLR = ONEWIRE_BIT; // Set pin low
  nrf_delay_us(600);              // Hold bus low for at least 480uS

  // Release bus and switch to input until presence pulse should be
  // visible.
  dirclr();          // Set pin as input
  nrf_delay_us(180); //  Eric - Wait at least 60uS before reading the bus.
                     // Read the bus in the middle of the cycle (60 +120).

  // Record presence if bus is low (DS182x is holding it there)
  present = !!(NRF_GPIO->IN & ONEWIRE_BIT); // Read the pin and invert. Present = 1 means the device present.

  // Wait for reset cycle to complete
  nrf_delay_us(300); //  Eric wait the full 480uS to make sure the DS18B20 is done.

  return present;
}

void vBSPACMonewireWriteByte_ni(int byte) {
  int bp;
  dirclr();        // Make sure bus is high to begin with. Setting pin as input so pullup sets to 1
  nrf_delay_us(5); // All we need is 1uS or so.

  for (bp = 0; bp < 8; ++bp) {
    NRF_GPIO->OUTCLR = ONEWIRE_BIT; // Set pin output low
    dirset();                       // now, set pin as output to drive bus low.
    if (byte & 0x01) {
      nrf_delay_us(8);  // Write a 1
      dirclr();         // Release the bus by setting pin as input so pullup sets to 1
      nrf_delay_us(60); // hold till end of write 1 cycle.
    } else {            // Write a 0
      nrf_delay_us(60); // Hold bus low to write 0
      dirclr();         // Release the bus by setting pin as input so pullup sets to 1
    }
    byte >>= 1;
    nrf_delay_us(5); // Min recovery time > 2uS before next bit.
  }
}

int iBSPACMonewireReadBit_ni(void) {
  int rv;
  dirclr();                          // Make sure bus is high to begin with.
                                     // Set pin as input so pullup sets to 1
  nrf_delay_us(5);                   // All we need is a minimum of 1uS
  NRF_GPIO->OUTCLR = ONEWIRE_BIT;    // Set pin output low
  dirset();                          // set pin as output to drive low.
  nrf_delay_us(5);                   // keep low for a min of 1uS
  dirclr();                          // release bus to read bit.
  nrf_delay_us(8);                   // wait till ready for read
  rv = (NRF_GPIO->IN & ONEWIRE_BIT); // read the value
  nrf_delay_us(45);                  // wait for device to release bus

  return rv;
}

int iBSPACMonewireReadByte_ni(void) {
  int byte = 0;
  int bit = 1;

  do {
    if (iBSPACMonewireReadBit_ni()) {
      byte |= bit;
    }
    bit <<= 1;
  } while (0x100 != bit);
  return byte;
}

int perform_temp_conversion(void) {
  if (!iBSPACMonewireReset_ni()) {
    return -1;
  }
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_CONVERT_T);
  dirclr();   // return ONEWIRE_PIN to input
  //  Go and wait for timer signaling that Dallas sensor competed conversion   
  return 0;
}

int read_temp_probe(int *temp_xCel) {
  int t;

  if (!iBSPACMonewireReset_ni()) {
    return -1;
  }
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
  nrf_delay_us(25); // delay for testing purposes
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_READ_SCRATCHPAD);
  nrf_delay_us(25); // delay for testing purposes
  t = iBSPACMonewireReadByte_ni();
  nrf_delay_us(25); // delay for testing purposes
  t |= (iBSPACMonewireReadByte_ni() << 8);

  return t;
}

/******************************************************************/
//            End of Temperature sensor module
// ****************************************************************
/**********************************************************************
************* PWM code for Audio out  *********************************
***********************************************************************/
APP_PWM_INSTANCE(PWM1,2);                   // Create the instance "PWM1" using TIMER 2. Timer 1 used by app_timers
static volatile bool ready_flag;            // A flag indicating PWM status.
void pwm_ready_callback(uint32_t pwm_id)    // PWM callback function
{
    ready_flag = true;
}

int pwm_init(void)
{
    ret_code_t err_code;

    // 1-channel PWM - Frequency setting (value is inverse frequency. 1000=1Khz)
    app_pwm_config_t pwm1_cfg = APP_PWM_DEFAULT_CONFIG_1CH(440L, AUDIO_OUT);

    // Initialize and enable PWM. 
    err_code = app_pwm_init(&PWM1,&pwm1_cfg,pwm_ready_callback);
    APP_ERROR_CHECK(err_code);
    app_pwm_enable(&PWM1);

    uint32_t pwm_value;
      pwm_value = 50;           // Duty cycle - was 80
      ready_flag = false;
            // Set the duty cycle - keep trying until PWM is ready... 
      while (app_pwm_channel_duty_set(&PWM1, 0, pwm_value) == NRF_ERROR_BUSY);

            // ... or wait for callback. 
      while (!ready_flag);
}

/*******************************************************************
/*************************** Beeper Routine ************************
/*******************************************************************/
void beep(uint32_t num_of_beeps, uint32_t time_on_ms, uint32_t time_off_ms) 
{
    do
    {
      nrf_gpio_cfg_output(AUDIO_OUT);
      nrf_delay_ms(time_on_ms);
      NRF_GPIO->PIN_CNF[AUDIO_OUT] = 0 | (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos) 
                                     | (GPIO_PIN_CNF_DRIVE_S0D1 << GPIO_PIN_CNF_DRIVE_Pos); 
      nrf_delay_ms(time_off_ms);
      num_of_beeps --;
    }while (num_of_beeps != 0);

//      nrf_gpio_pin_clear(AUDIO_OUT);         // Set pin low - turn off for debugging
}

/*************************** End Beeper Routine *********************/

/*********************************************************************
*                   Stepper Motor Control
**********************************************************************/

// Stepper Motor Step sequiences for STEP_A & STEP_B GPIOs
// #define STEP_CLOCKWISE[4] {STATE_0 = 00, STATE_1 = 10, STATE_2 = 11, STATE_3 = 01}  
// #define STEP_COUNTERCLOCKWISE[4] {STATE_0 = 00, STATE_3 = 01, STATE_2 = 11, STATE_1 = 10}  

  // Assume that the valve adjustment function takes less time than the temperature conversion 
  // timer of ~195ms so we don't have the temperature changing during an adjustment cycle.
  // If this does happen, we might have to rethink this code.

void adjust_temp(int target_temp)   
{
//  while(1)
//  {
  // ****************** Enable the stepper motor - Reduces motor heating during testing *****************
//  nrf_gpio_pin_set(TEMP_OVERRIDE);
//  nrf_delay_ms(1); // Give motor driver some time.
  // ************************** End *********************************************************************

  // This is where we adjust the valve to get the actual water temperature to match the desired temperature
  // This is the main control loop that we stay in until a timer expires or until the wireless controller 
  // changes something.

int delta_t = abs(target_temp - t_avg);     // determine how far off the water temperature is from target
int rate_of_change = abs(present_temp - t_avg);

  if (target_temp > t_avg)                  // we want to heat
  {
    heat = true;                            // heat varaible is used in rotate_valve routine
//    stepper_motor_rate = 10;
//    rotate_valve(delta_t);                  // We only start to this routine via purge valve.

    // Determine the water temp change rate and direction
    // Water temp can change much faster than our temperature sensor can respond so we have to go slow
    if (present_temp >= t_avg)               // we're heating
    {
      if ((rate_of_change >= 10) && (present_temp > target_temp - rate_of_change) && (present_temp <= target_temp)) 
      {
//        stepper_motor_rate = 30;              // we can get one to two new temperature readings during fast changeing water 
        heat = false;                         // start reversing the valve to close in on the target temp
        rotate_valve(rate_of_change/1.8);     // adjust valve one degree cooler per degree of rate of change.
        heat = true;  
      }
      if ((rate_of_change >= 3) && (present_temp > target_temp - rate_of_change) && (present_temp <= target_temp)) // this can happen when we purge the valve
      {
//        stepper_motor_rate = 70;              // we can get one to two new temperature readings during fast changeing water 
        heat = false;                         // start reversing the valve to close in on the target temp
        rotate_valve(rate_of_change/1.8);     // adjust valve one degree cooler per degree of rate of change.
        heat = true;  

      } else if (rate_of_change >= 0 && rate_of_change < 3)  // rate of change is slowing
      {
//        stepper_motor_rate = 200;          // use a slower stepper rate to control the rate of change more accurately
        if (t_avg < target_temp)
        {
          heat = true;  
          rotate_valve(1);               // haven't reached the desired temp yet add heat
        }else if (t_avg > target_temp)
        {  
          heat = false;                  // start reversing the valve to close in on the target temp
          rotate_valve(1);
          heat = true;  
        }
      } // Let it heat                              
    }
  } else
  { 
  if (target_temp < t_avg)                  // we want to cool
   {
    heat = false;                            // heat varaible is used in rotate_valve routine
//    stepper_motor_rate = 10;
//    rotate_valve(delta_t);                  // We only start to this routine via purge valve.

    // Determine the water temp change rate and direction
    // Water temp can change much faster than our temperature sensor can respond so we have to go slow
    if (present_temp <= t_avg)               // we're cooling
    {
      if ((rate_of_change >= 10) && (present_temp < target_temp + rate_of_change) && (present_temp >= target_temp)) 
      {
//        stepper_motor_rate = 30;              // we can get one to two new temperature readings during fast changeing water 
        heat = true;                         // start reversing the valve to close in on the target temp
        rotate_valve(rate_of_change/1.8);     // adjust valve one degree hotter per degree of rate of change.
        heat = false;  
      }
      if ((rate_of_change >= 3) && (present_temp < target_temp + rate_of_change) && (present_temp >= target_temp)) // this can happen when we purge the valve
      {
//        stepper_motor_rate = 70;              // we can get one to two new temperature readings during fast changeing water 
        heat = true;                         // start reversing the valve to close in on the target temp
        rotate_valve(rate_of_change/1.8);     // adjust valve one degree hotter per degree of rate of change.
        heat = false;  

      } else if (rate_of_change >= 0 && rate_of_change < 3)  // rate of change is slowing
      {
//        stepper_motor_rate = 200;          // use a slower stepper rate to control the rate of change more accurately
        if (t_avg < target_temp)
        {
          heat = true;                  // start reversing the valve to close in on the target temp
          rotate_valve(1);
          heat = false;  
        }else if (t_avg > target_temp)
        {  
          heat = false;                  // start reversing the valve to close in on the target temp
          rotate_valve(1);
        }
      } // Let it cool                              
    }
   }
  }
}

//************************ Step Forward **********************/
void step_forward()
{
      if (stepper_state == 0) 
      {
        nrf_gpio_pin_set(STEP_B);
        nrf_gpio_pin_clear(STEP_A);
        stepper_state = 1;
        start_stepper_timer();
        goto done_forward;
      }
      if (stepper_state == 1)
      {
        nrf_gpio_pin_set(STEP_B);
        nrf_gpio_pin_set(STEP_A);
        stepper_state = 2;
        start_stepper_timer();
        goto done_forward;
      } 
      if (stepper_state == 2)
      {
        nrf_gpio_pin_set(STEP_A);
        nrf_gpio_pin_clear(STEP_B);
        stepper_state = 3;
        start_stepper_timer();
        goto done_forward;
      } 
      if (stepper_state == 3)
      {
        nrf_gpio_pin_clear(STEP_A);
        nrf_gpio_pin_clear(STEP_B);
        stepper_state = 0;
        start_stepper_timer();
        goto done_forward;
      }
start_stepper_timer();

done_forward:;
}
//************************ Step Backward **********************/
void step_backward()
{
      if (stepper_state == 0)
      {                             
        nrf_gpio_pin_set(STEP_A);
        nrf_gpio_pin_clear(STEP_B);
        stepper_state = 3;
        start_stepper_timer();
        goto done_backward;
      } 
      if (stepper_state == 1)
      {
        nrf_gpio_pin_clear(STEP_A);
        nrf_gpio_pin_clear(STEP_B);
        stepper_state = 0;
        start_stepper_timer();
        goto done_backward;
      }
      if (stepper_state == 2)
      {
        nrf_gpio_pin_set(STEP_B);
        nrf_gpio_pin_clear(STEP_A);
        stepper_state = 1;
        start_stepper_timer();
        goto done_backward;     
      }  
      if (stepper_state == 3)
      {
        nrf_gpio_pin_set(STEP_A);
        nrf_gpio_pin_set(STEP_B);
        stepper_state = 2;
        start_stepper_timer();
        goto done_backward;
      }
start_stepper_timer();

done_backward:;
}

/*
// ****************** Disable the stepper motor - Reduces motor heating during testing *****************
//  nrf_gpio_pin_clear(TEMP_OVERRIDE);
//  nrf_delay_ms(1); // Give motor driver some time.
  // ************************** End *********************************************************************
*/
// End of function


static bool turning_valve = false;
static volatile uint8_t do_steps;

void rotate_valve(uint8_t num_steps)
{

    if (!turning_valve)
    {
      do_steps = num_steps;
      turning_valve = true;
    }
  //    nrf_gpio_pin_set(TEMP_OVERRIDE);  // enable stepper driver. Temp at desired_temp

    while(do_steps != 0)
    {
//      uint8_t do_steps;    // determine which way we need to step in order to heat or cool
      if (clockwise_heat && heat)
      {
        stepping_forward = true;
        step_forward();              
         do_steps --;
         goto stepper_wait;
      } else if (clockwise_heat && !heat)
      {
        stepping_forward = false;
        step_backward();              
        do_steps --;
        goto stepper_wait;
      } else if (!clockwise_heat && heat)
      {
        stepping_forward = false;
        step_backward();              
        do_steps --;
        goto stepper_wait;
      } else if (!clockwise_heat && !heat)
      {
        stepping_forward = true;
        step_forward();              
        do_steps --;
        goto stepper_wait;
      }
    }
    turning_valve = false;

stepper_wait:;
}
/*********************************************************************
*                                                                    *
*                 ADD CODE TO TURN OFF VALVE HERE                    *
*                                                                    *
**********************************************************************/

/*********************************************************************
*                    End of Stepper Motor Control
**********************************************************************/

/*****************************************************************************************
                ADC code.  Modified to read system voltage on P0.30/AIN6
******************************************************************************************/
int get_sys_v(void){

// figure out this code.

}               

void saadc_sampling_event_init(void) {
  ret_code_t err_code;

//  err_code = nrf_drv_ppi_init();
//  APP_ERROR_CHECK(err_code);

  nrf_drv_timer_config_t timer_cfg = NRF_DRV_TIMER_DEFAULT_CONFIG;
  timer_cfg.bit_width = NRF_TIMER_BIT_WIDTH_32;
  err_code = nrf_drv_timer_init(&m_timer, &timer_cfg, saadc_timer_handler);
  APP_ERROR_CHECK(err_code);

  //    setup m_timer for compare event every 400ms
  uint32_t ticks = nrf_drv_timer_ms_to_ticks(&m_timer, 400);
  nrf_drv_timer_extended_compare(&m_timer,
      NRF_TIMER_CC_CHANNEL0,
      ticks,
      NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK,
      false);
  nrf_drv_timer_enable(&m_timer);

  uint32_t timer_compare_event_addr = nrf_drv_timer_compare_event_address_get(&m_timer,
      NRF_TIMER_CC_CHANNEL0);
  uint32_t saadc_sample_task_addr = nrf_drv_saadc_sample_task_get();

  //    setup ppi channel so that timer compare event is triggering sample task in SAADC
  err_code = nrf_drv_ppi_channel_alloc(&m_ppi_channel);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_ppi_channel_assign(m_ppi_channel,
      timer_compare_event_addr,
      saadc_sample_task_addr);
  APP_ERROR_CHECK(err_code);
}

void saadc_sampling_event_enable(void) {
  ret_code_t err_code = nrf_drv_ppi_channel_enable(m_ppi_channel);
  APP_ERROR_CHECK(err_code);
}

void saadc_callback(nrf_drv_saadc_evt_t const *p_event) {

  if (p_event->type == NRF_DRV_SAADC_EVT_DONE) {
    ret_code_t err_code;

    err_code = nrf_drv_saadc_buffer_convert(p_event->data.done.p_buffer, SAMPLES_IN_BUFFER);
    APP_ERROR_CHECK(err_code);

    int i;
    int bat_level;

    NRF_LOG_INFO("ADC event number: %d", (int)m_adc_evt_counter);

    for (i = 0; i < SAMPLES_IN_BUFFER; i++) {

    bat_level = p_event->data.done.p_buffer[i]>>2;
//  At this point your battery level is captured in bat_level. You can do whatever you want with it now

      NRF_LOG_INFO("Battery Level = %d", p_event->data.done.p_buffer[i]);
    }
    m_adc_evt_counter++;
  }
}

void saadc_init(void) {
  ret_code_t err_code;
  nrf_saadc_channel_config_t channel_config =
  NRF_DRV_SAADC_DEFAULT_CHANNEL_CONFIG_SE(NRF_SAADC_INPUT_AIN6); // PS uses SYS_V (Pin 30-AIN6) for Sys_Pwr monitoring

  err_code = nrf_drv_saadc_init(NULL, saadc_callback);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_channel_init(0, &channel_config);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[0], SAMPLES_IN_BUFFER);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[1], SAMPLES_IN_BUFFER);
  APP_ERROR_CHECK(err_code);
  // We only want to sample when we need the voltage info... Not continually

  saadc_sampling_event_enable();
  NRF_LOG_INFO("SAADC HAL simple example started.");

/*************** Turn off SAADC ******************************
nrf_drv_saadc_uninit(); //Turns off SAADC
    NRF_SAADC->INTENCLR = (SAADC_INTENCLR_END_Clear << SAADC_INTENCLR_END_Pos);
    NVIC_ClearPendingIRQ(SAADC_IRQn);
*/
  return;
}

/**************************************************************************
********************* Sample code to read ADC output **********************
***************************************************************************/

/**@brief ADC interrupt handler.
 * @details  This function will fetch the conversion result from the ADC, convert the value into
 *           percentage and send it to peer.
 */
 /*
void ADC_IRQHandler(void)
{
    if (NRF_ADC->EVENTS_END != 0)
    {
        uint8_t     adc_result;
        uint16_t    batt_lvl_in_milli_volts;
        uint8_t     batt_lvl_formatted;			
        uint8_t     percentage_batt_lvl;
        uint32_t    err_code;

        NRF_ADC->EVENTS_END     = 0;
        adc_result              = NRF_ADC->RESULT;
        NRF_ADC->TASKS_STOP     = 1;

        batt_lvl_in_milli_volts = ADC_RESULT_IN_MILLI_VOLTS(adc_result) ;//+ DIODE_FWD_VOLT_DROP_MILLIVOLTS;
        percentage_batt_lvl     = battery_level_in_percent(batt_lvl_in_milli_volts);
				
			  batt_lvl_formatted = batt_lvl_in_milli_volts/100;
			
			  if (batt_lvl_formatted != m_bas.battery_level_last)
        {
					uint16_t len = sizeof(uint8_t);
					
					// Save new battery value
					m_bas.battery_level_last = batt_lvl_formatted;
					
					// Update database
					err_code = sd_ble_gatts_value_set(m_bas.battery_level_handles.value_handle,
																						0,
																						&len,
																						&batt_lvl_formatted);					
				}
        err_code = ble_bas_battery_level_update(&m_bas, percentage_batt_lvl);
        if (
            (err_code != NRF_SUCCESS)
            &&
            (err_code != NRF_ERROR_INVALID_STATE)
            &&
            (err_code != BLE_ERROR_NO_TX_BUFFERS)
            &&
            (err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING)
        )
        {
            APP_ERROR_HANDLER(err_code);
        }
    }
}


/**@brief Function to make the ADC start a battery level conversion.
 */
 /*
static void adc_start(void)
{
    uint32_t err_code;

    // Configure ADC
    NRF_ADC->INTENSET   = ADC_INTENSET_END_Msk;
    NRF_ADC->CONFIG     = (ADC_CONFIG_RES_8bit                        << ADC_CONFIG_RES_Pos)     |
                          (ADC_CONFIG_INPSEL_SupplyOneThirdPrescaling << ADC_CONFIG_INPSEL_Pos)  |
                          (ADC_CONFIG_REFSEL_VBG                      << ADC_CONFIG_REFSEL_Pos)  |
                          (ADC_CONFIG_PSEL_Disabled                   << ADC_CONFIG_PSEL_Pos)    |
                          (ADC_CONFIG_EXTREFSEL_None                  << ADC_CONFIG_EXTREFSEL_Pos);
    NRF_ADC->EVENTS_END = 0;
    NRF_ADC->ENABLE     = ADC_ENABLE_ENABLE_Enabled;

    // Enable ADC interrupt
    err_code = sd_nvic_ClearPendingIRQ(ADC_IRQn);
    APP_ERROR_CHECK(err_code);

    err_code = sd_nvic_SetPriority(ADC_IRQn, NRF_APP_PRIORITY_LOW);
    APP_ERROR_CHECK(err_code);

    err_code = sd_nvic_EnableIRQ(ADC_IRQn);
    APP_ERROR_CHECK(err_code);

    NRF_ADC->EVENTS_END  = 0;    // Stop any running conversions.
    NRF_ADC->TASKS_START = 1;		
}
*/


/*************************** More "Bare_Metal" code from Github ********************************\
#include <nrf.h>

int mian(void)
{
  volatile int16_t result = 0;
  volatile float precise_result = 0;

  // Start HFCLK from crystal oscillator, this will give the SAADC higher accuracy
  NRF_CLOCK->TASKS_HFCLKSTART = 1;
  while (NRF_CLOCK->EVENTS_HFCLKSTARTED == 0);
  NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;

  // Configure SAADC singled-ended channel, Internal reference (0.6V) and 1/6 gain.
  NRF_SAADC->CH[0].CONFIG = (SAADC_CH_CONFIG_GAIN_Gain1_6    << SAADC_CH_CONFIG_GAIN_Pos) |
                            (SAADC_CH_CONFIG_MODE_SE         << SAADC_CH_CONFIG_MODE_Pos) |
                            (SAADC_CH_CONFIG_REFSEL_Internal << SAADC_CH_CONFIG_REFSEL_Pos) |
                            (SAADC_CH_CONFIG_RESN_Bypass     << SAADC_CH_CONFIG_RESN_Pos) |
                            (SAADC_CH_CONFIG_RESP_Bypass     << SAADC_CH_CONFIG_RESP_Pos) |
                            (SAADC_CH_CONFIG_TACQ_3us        << SAADC_CH_CONFIG_TACQ_Pos);

  // Configure the SAADC channel with VDD as positive input, no negative input(single ended).
  NRF_SAADC->CH[0].PSELP = SAADC_CH_PSELP_PSELP_VDD << SAADC_CH_PSELP_PSELP_Pos;
  NRF_SAADC->CH[0].PSELN = SAADC_CH_PSELN_PSELN_NC << SAADC_CH_PSELN_PSELN_Pos;

  // Configure the SAADC resolution.
  NRF_SAADC->RESOLUTION = SAADC_RESOLUTION_VAL_14bit << SAADC_RESOLUTION_VAL_Pos;

  // Configure result to be put in RAM at the location of "result" variable.
  NRF_SAADC->RESULT.MAXCNT = 1;
  NRF_SAADC->RESULT.PTR = (uint32_t)&result;

  // No automatic sampling, will trigger with TASKS_SAMPLE.
  NRF_SAADC->SAMPLERATE = SAADC_SAMPLERATE_MODE_Task << SAADC_SAMPLERATE_MODE_Pos;

  // Enable SAADC (would capture analog pins if they were used in CH[0].PSELP)
  NRF_SAADC->ENABLE = SAADC_ENABLE_ENABLE_Enabled << SAADC_ENABLE_ENABLE_Pos;

  // Calibrate the SAADC (only needs to be done once in a while)
  NRF_SAADC->TASKS_CALIBRATEOFFSET = 1;
  while (NRF_SAADC->EVENTS_CALIBRATEDONE == 0);
  NRF_SAADC->EVENTS_CALIBRATEDONE = 0;
  while (NRF_SAADC->STATUS == (SAADC_STATUS_STATUS_Busy <<SAADC_STATUS_STATUS_Pos));

  // Start the SAADC and wait for the started event.
  NRF_SAADC->TASKS_START = 1;
  while (NRF_SAADC->EVENTS_STARTED == 0);
  NRF_SAADC->EVENTS_STARTED = 0;

  // Do a SAADC sample, will put the result in the configured RAM buffer.
  NRF_SAADC->TASKS_SAMPLE = 1;
  while (NRF_SAADC->EVENTS_END == 0);
  NRF_SAADC->EVENTS_END = 0;

  // Convert the result to voltage
  // Result = [V(p) - V(n)] * GAIN/REFERENCE * 2^(RESOLUTION)
  // Result = (VDD - 0) * ((1/6) / 0.6) * 2^14
  // VDD = Result / 4551.1
  precise_result = (float)result / 4551.1f;
  precise_result; // to get rid of set but not used warning

  // Stop the SAADC, since it's not used anymore.
  NRF_SAADC->TASKS_STOP = 1;
  while (NRF_SAADC->EVENTS_STOPPED == 0);
  NRF_SAADC->EVENTS_STOPPED = 0;


  while (1)
  {
    __WFE();
  }
}
//********************************************************************
//                        End SAADC code
//********************************************************************

/*****************************************************************************************
                     Flash Routines
******************************************************************************************/
//  flash_operation_completed = false;

    // Set flash location
uint32_t static volatile pg_size;
uint32_t static volatile pg_num;
uint32_t static volatile flash_addr;

static uint32_t nrf5_flash_end_addr_get()
{
    pg_size = NRF_FICR->CODEPAGESIZE;
    pg_num = NRF_FICR->CODESIZE-2; // Leave room for BLE stuff
    flash_addr = (uint32_t *)(pg_size * pg_num);
}


/****************** Erase Flash ***************************/
void erase_flash_page(void)
{
    NRF_LOG_INFO("Erasing Flash...");
    uint32_t evt = 0;
retry_flash:
    flash_operation_completed = false;
    sd_flash_page_erase(pg_num);
/*   
    while(!flash_operation_completed)
    {
      sd_evt_get (&evt);
      if (evt == NRF_EVT_FLASH_OPERATION_SUCCESS || evt == NRF_EVT_FLASH_OPERATION_ERROR)
      {
        flash_operation_completed = true;
        goto resume_flash_operations;
      }
      else 
      {
        flash_operation_completed = true;
        start_general_timer();
        goto retry_flash;
      }
    }
*/
    while(!flash_operation_completed)
    {
//    nrf_delay_ms(100);     // 85ms minimum flash time to finish erase cycle
//    sd_app_evt_wait();
    flash_operation_completed = true;
    }

resume_flash_operations:
    return (12345);
}

/****************** Write Flash ***************************/
void write_flash_page(void)
{
    NRF_LOG_INFO("Writing Flash...");
    uint8_t block_size = sizeof(write_flash_block);
    ret_code_t err_code;
    flash_operation_completed = false;
    err_code = sd_flash_write((uint32_t*)flash_addr, (uint32_t*)&write_flash_block, block_size);

//    while(!flash_operation_completed);
/*
    if (err_code != NRF_SUCCESS)
    {
       NRF_LOG_INFO("Error writing to flash, err_code = \"%x\" .", err_code);
        if (err_code == NRF_ERROR_INVALID_LENGTH) while (1);
        if (err_code == NRF_ERROR_INVALID_ADDR) while (1);
        if (err_code == NRF_ERROR_BUSY) while (1);
        if (err_code == NRF_ERROR_FORBIDDEN) while (1);
    }
*/

/*    while(!flash_operation_completed)
    {
    sd_app_evt_wait();
    nrf_delay_ms(2);     // <1 ms minimum flash time to finish write cycle
    flash_operation_completed = true;
    }
*/
}

/******************** Read Flash ****************************/
void get_flash_data(void)
{
    uint8_t block_size = sizeof(read_flash_block);
    memcpy((uint32_t*)&read_flash_block, (uint32_t*)flash_addr, block_size);
}


int flash_init(void)
{
    // Get our flash page info
    nrf5_flash_end_addr_get();
    // Read flash settings into read_flash_block structure
    get_flash_data();

    if (flash_test)
    {
    // Now print it out
    NRF_LOG_INFO("Read back flash data desired_temp \"%d\" from flash.", read_flash_block.desired_temp);
    NRF_LOG_INFO("Read back flash data desired_flow \"%d\" from flash.", read_flash_block.desired_flow);
    NRF_LOG_INFO("Read back flash data user_max_temp \"%d\" from flash.", read_flash_block.user_max_temp);
    NRF_LOG_INFO("Read back flash data absolute_max_temp \"%d\" from flash.", read_flash_block.absolute_max_temp);
    NRF_LOG_INFO("Read back flash data startup_done \"%x\" from flash.", read_flash_block.startup_done);
    NRF_LOG_INFO("Read back flash data install_mode \"%x\" from flash.", read_flash_block.install_mode);
    NRF_LOG_INFO("Read back flash data initialization_done \"%x\" from flash.", read_flash_block.initialization_done);
    NRF_LOG_INFO("Read back flash data clockwise_heat \"%x\" from flash.", read_flash_block.clockwise_heat);
    NRF_LOG_INFO("Read back flash data shower_timer \"%d\" from flash.", read_flash_block.shower_timer);
    NRF_LOG_INFO("Read back flash data temp_scale \"%x\" from flash.", read_flash_block.temp_scale);
    NRF_LOG_INFO("Read back flash data user_last_preset \"%d\" from flash.", read_flash_block.user_last_preset);
    NRF_LOG_INFO("Read back flash data user_last_temp \"%d\" from flash.", read_flash_block.user_last_temp);
    NRF_LOG_INFO("Read back flash data flash_check \"%x\" from flash.", read_flash_block.flash_check);
    }
}

void update_flash_info()
{
    if (update_flash)     // fill flash write structure with updated system values
    {
      write_flash_block.desired_temp = desired_temp;                
      write_flash_block.desired_flow = desired_flow;    
      write_flash_block.user_max_temp = user_max_temp;              
      write_flash_block.absolute_max_temp = absolute_max_temp;     
      write_flash_block.startup_done = startup_done;
      write_flash_block.install_mode = install_mode;                
      write_flash_block.initialization_done = initialization_done;  
      write_flash_block.clockwise_heat = clockwise_heat;            
      write_flash_block.shower_timer = shower_timer;                
      write_flash_block.temp_scale = temp_scale;                    
      write_flash_block.user_last_preset = user_last_preset;
      write_flash_block.user_last_temp = user_last_temp;
      write_flash_block.flash_check = 0x55AA55AA;

      erase_flash_page();
      nrf_delay_ms(200);
      write_flash_page();
      nrf_delay_ms(200);
      get_flash_data();
      update_flash = false;
//      nrf_delay_ms(500);

//      sd_nvic_SystemReset();
      perfect_shower_init();
//      sysResetReq();
//      sd_softdevice_disable();
//      nrf_sdh_disable_request();
//      sd_nvic_SystemReset();
//      perfect_shower_init();
//      NVIC_SystemReset();
//void forceSysReset(void) { ((void (code *) (void)) 0x0000) (); }
    }
}

/*****************************************************************************************
                      END- Flash Routines
******************************************************************************************/
