/**
 * Copyright (c) 2016 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "app_error.h"
#include "app_gpiote.h"
#include "app_pwm.h"
#include "app_timer.h"
#include "app_uart.h"
#include "app_util.h"
#include "app_util_platform.h"
#include "ble.h"
#include "ble_db_discovery.h"
#include "ble_gap.h"
#include "ble_hci.h"
#include "ble_nus_c.h"
#include "boards.h"
#include "bsp.h"
#include "bsp_btn_ble.h"
#include "fds.h"
//#include "fds_example.h"
#include "low_power_pwm.h"
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_scan.h"
#include "nrf_cli.h"
#include "nrf_cli_uart.h"
#include "nrf_delay.h"
#include "nrf_drv_clock.h"
#include "nrf_drv_gpiote.h"
#include "nrf_drv_ppi.h"
#include "nrf_drv_pwm.h"
#include "nrf_drv_saadc.h"
#include "nrf_drv_timer.h"
#include "nrf_drv_uart.h"
#include "nrf_gpio.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_log_internal.h"
#include "nrf_nvmc.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_queue.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "nrfx_timer.h"
#include "nrfx_pwm.h"
#include "sdk_errors.h"
#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#define NRF_LOG_MODULE_NAME app
#define NRF_CLI_ENABLED 1

#define APP_BLE_CONN_CFG_TAG 1  /**< Tag that refers to the BLE stack configuration set with @ref sd_ble_cfg_set. The default tag is @ref BLE_CONN_CFG_TAG_DEFAULT. */
#define APP_BLE_OBSERVER_PRIO 3 /**< BLE observer priority of the application. There is no need to modify this value. */

#define UART_TX_BUF_SIZE 256 /**< UART TX buffer size. */
#define UART_RX_BUF_SIZE 256 /**< UART RX buffer size. */

#define NUS_SERVICE_UUID_TYPE BLE_UUID_TYPE_VENDOR_BEGIN /**< UUID type for the Nordic UART Service (vendor specific). */

/******************************* DFH *************************/
#define ECHOBACK_BLE_UART_DATA 0 /**< Echo the UART data that is received over the Nordic UART Service (NUS) back to the sender. */
/******************************* DFH *************************/

#define Board_PS false       // Change to true for Perfect Shower Valve build, false for nRF52DK development board
#define Board_52DK !Board_PS // Used for debugging on nRF52DK system

// Pin Definitions for Perfect Shower RevD (MS50SFB) PCBA
#if Board_PS
// nRF52832 GPIO#                 // 52832 Pin# - Port0 # - MS50SFB Pin#
#define STEP_A 28        // pin #40 - P0_28/AIN4 - 6
#define STEP_B 29        // pin #41 - P0_29/AIN5 - 7
#define LED_1 17         // pin #20 - P0.17      - 32 Grn
#define LED_2 18         // pin #21 - P0.18      - 33 Red
#define SYS_V 30         // pin #42 - P0.30/AIN6 - 8
#define A_TEMP_OUT 31    // pin #43 - P0.31/AIN7 - 9
#define A_TEMP_IN 2      // pin #4  - P0.02/AIN0 - 16
#define AUDIO_OUT 3      // pin #5  - P0.03/AIN1 - 17 (buzzer)
#define MAN_OVERRIDE 11  // pin #14 - P0.11      - 26
#define VALVE_OFF 12     // pin #15 - P0.12      - 27
#define MAN_RESET 21     // pin #24 - P0.21/RST  - 36
#define TEMP_OVERRIDE 22 // pin #27 - P0.22      - 39
#define D_TEMP 23        // pin #28 - P0.23      - 40
#define ONEWIRE_PIN 23   // def for onewire code
#define TRIP_TEST 24     // pin #29 - P0.24      - 41
#define d_temp_irq 25    // pin #37 - P0.25      - P4
#endif

/* Pin Definitions for nRF52DK system */
/* nRF52DK GPIO#                  // 52832 Pin # - Port0.# - DK Hdr */
#if Board_52DK
#define LED_1 17         // pin #20 - P0.17 LED_1 - P3 Grn
#define LED_2 18         // pin #21 - P0.18 LED_2 - P3 Grn
#define STEP_A 28        // pin #40 - P0_28/AIN4  - P2   // Eric was 28
#define STEP_B 29        // pin #41 - P0_29/AIN5  - P2
#define SYS_V 30         // pin #42 - P0.30/AIN6  - P2
#define A_TEMP_OUT 31    // pin #43 - P0.31/AIN7  - P2
#define A_TEMP_IN 2      // pin #4  - P0.02/AIN0  - P4
#define AUDIO_OUT 3      // pin #5  - P0.03/AIN1  - P2 (buzzer)
#define MAN_OVERRIDE 11  // pin #14 - P0.11       - P3
#define VALVE_OFF 12     // pin #15 - P0.12       - P3
#define MAN_RESET 21     // pin #24 - P0.21/RST   - P6
#define TEMP_OVERRIDE 22 // pin #27 - P0.22       - P4
#define D_TEMP 23        // pin #28 - P0.23       - P4
#define ONEWIRE_PIN 23   // def for onewire code         // Eric was 23
#define TRIP_TEST 24     // pin #29 - P0.24       - P4
#define D_TEMP_IRQ 25    // pin 37  - P0.25       - P4
#endif
/*
// Unused pins
#define P0_26           26        // pin #38
#define P0_27           27        // pin #39
#define P0_00_XAL1      0         // pin #2
#define P0_01_XAL2      1         // pin #3
#define P0_04           4         // pin #6
#define P0_05_AIN3      5         // pin #7
#define P0_06           6         // pin #8
#define P0_07           7         // pin #9
#define P0_08           8         // pin #10
#define P0_09_NFC1      9         // pin #11
#define P0_10_NFC2      10        // pin #12
#define P0_11           11        // pin #14
#define P0_13           13        // pin #16
#define P0_14           14        // pin #17
#define P0_15           15        // pin #18
#define P0_16           16        // pin #19
#define P0_19           19        // pin #22
#define P0_20           20        // pin #23
*/

#define LEDS_NUMBER 2
#define LEDS_ACTIVE_STATE 0

#define LEDS_INV_MASK LEDS_MASK
#define LEDS_LIST { LED_1, LED_2 }
#define BSP_LED_0 LED_1
#define BSP_LED_1 LED_2

#define RX_PIN_NUMBER 8
#define TX_PIN_NUMBER 6
#define CTS_PIN_NUMBER 7
#define RTS_PIN_NUMBER 5
#define HWFC true

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

#define HOUSEKEEPING_TIMER 1500 /**< (app_timer_ticks) Run perfect_shower_main code for housekeeping. ~1.5 sec */
#define STEPPER_MOTOR_RATE 50   /**< (app_timer_ticks) Stepper motor step rate ~50ms per step */

#define ECHOBACK_BLE_UART_DATA 0 /**< Echo the UART data that is received over the Nordic UART Service (NUS) back to the sender. */
#define SETSDATA(A, B, C, D, E, F) \
  sdata[0] = A;                    \
  sdata[1] = B;                    \
  sdata[2] = C;                    \
  sdata[3] = D;                    \
  sdata[4] = E;                    \
  sdata[5] = F;

// #define NRFX_SYSTICK_ENABLED  SYSTICK_ENABLED         // Eric
// #define NRFX_TIMER1_ENABLED  TIMER1_ENABLED           // Eric enabled in sdk_config.h as well

uint32_t sys_timer;
static bool temp_scale = false;               // F/C. False=Farenheit, True=Celsius.  This value from user display
static uint32_t current_temp;                 // Global current water temperature
static uint32_t t_raw;                        // raw DS18B20 temperature sensor data
static  uint32_t t_count = 0;                 // temperature reading counter
static uint32_t desired_temp;                 // same thing?
static uint32_t tdata[5];                            // temperature storage for calculating average temp
static uint32_t t_avg = 0x1d;                 // average of 5 consecutive temperature readings (about 1 sec)
bool t_probe_present = false;                 // Temperature sensor is working
bool temp_probe_irq = false;                  // Tells us when a temperature conversion is complete
bool waiting_for_temp_conversion = false;     // Started a temp conversion cycle and waiting for it to finish
float t_dCel, t_dFahr;                        // Celsius and Farenheit temperature values
bool clockwise_heat = true;                   // To start, assume that water gets hotter with clockwise.
                                              // This will get changed by the initialization code and the
                                              // final value stored in flash for reading during power up.
                                              // Need to implement flash save!
bool purging_valve = false;                   // True only when we're purging the hot water.
// int32_t trendTemp1 = 0, trendTemp2 = 0;    // Temporary temperature value to determine the direction
// of the changing water temperature, i.e. hotter or colder
// byte tempReadCount = 0;                    // Keeps track of the number of sequential temperature
// readings for temperature trending info.
// bool trendingUp = false;                   // which way is the water temperature moving, up or down?
// int trendAmount = 0;                       // How fast is the water changing temperature?
int32_t absolute_max_temp = 0x1de;    // ?? value??  // need to use raw_temp not C or F to eliminate scale errors. 
                                              //  Default = 115 F. This is programmable by the user
                                              //    interface. This value will be saved in flash. Hardware
                                              //    safety temperature circuitry set to 120 F.
bool initialization_done = false;             // When the valve is first installed, the installation code
                                              //    will set this to "true".  This way we know all the system
                                              //    variables are configured properly

bool step_forward_timer_started = false;
bool step_backward_timer_started = false;
bool volatile install_mode = false;           // First time installation(& reset)mode.
bool volatile manual_mode = false;            // Need to know when the user is controlling things.
bool automatic_mode = false;                  // not manual mode
bool volatile program_mode = false;           // Used to know when we are recording a user programed profile.
bool volatile program_run_mode = false;       // Used to know when we are running a pre-programmed profile.
bool volatile reset_ps = false;               // If resetWG = true then set initializationDone and installMOde
                                              //   to false so we can reset the valve and perform
                                              //   installation agian.
int32_t system_voltage;
unsigned char valve_off_dir[] = "cool";       //  tells us which direction to turn off valve - "cool" or "heat"
int32_t users_hot_water_temp = 100;           //  measured during installation. Default scale = F
bool user_interface_connect = false;          // connectetd to wireless controller

/**********************************
 * Flash Memory Storage Structure *
 **********************************
*flash[1] = saved_Temp            *
*flash[2] = absoluteMaxTemp       *
*flash[3] = userMaxTemp           *
*flash[4] = clockwiseHeat         *
*flash[5] = install_mode           *
*flash[6] = InitializationDone    *
*flash[7] = bitResolution         *
*flash[8] =                       *
*flash[9] = 
*flash[10] = 
***********************************/

struct blockstruct // block to write into flash.
{
  int desired_temp; // set by user request, used to adjust valves.
  int desired_flow; // same
  int abs_max_temp;
  bool install_mode;
  bool initialization_done;
  bool clockwise_heat;
  int users_hot_water_temp;
  unsigned char valve_off_dir[];
} flashblock;

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

BLE_NUS_C_DEF(m_ble_nus_c);      /**< BLE Nordic UART Service (NUS) client instance. */
NRF_BLE_GATT_DEF(m_gatt);        /**< GATT module instance. */
BLE_DB_DISCOVERY_DEF(m_db_disc); /**< Database discovery module instance. */
NRF_BLE_SCAN_DEF(m_scan);        /**< Scanning Module instance. */

static uint16_t m_ble_nus_max_data_len = BLE_GATT_ATT_MTU_DEFAULT - OPCODE_LENGTH - HANDLE_LENGTH; /**< Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module. */

/**@brief NUS UUID. */
static ble_uuid_t const m_nus_uuid =
    {
        .uuid = BLE_UUID_NUS_SERVICE,
        .type = NUS_SERVICE_UUID_TYPE};

static char const m_target_periph_name[] = "WLC-01"; /**< Name of the device we try to connect to. This name is searched in the scan report data*/

/**@brief Function for handling asserts in the SoftDevice.
 *
 * @details This function is called in case of an assert in the SoftDevice.
 *
 * @warning This handler is only an example and is not meant for the final product. You need to analyze
 *          how your product is supposed to react in case of assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num     Line number of the failing assert call.
 * @param[in] p_file_name  File name of the failing assert call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t *p_file_name) {
  app_error_handler(0xDEADBEEF, line_num, p_file_name);
}

/**@brief Function for starting scanning. */
static void scan_start(void) {
  ret_code_t ret;

  ret = nrf_ble_scan_start(&m_scan);
  APP_ERROR_CHECK(ret);

  ret = bsp_indication_set(BSP_INDICATE_SCANNING);
  APP_ERROR_CHECK(ret);
}

/**@brief Function for handling Scanning Module events.
 */
static void scan_evt_handler(scan_evt_t const *p_scan_evt) {
  ret_code_t err_code;

  switch (p_scan_evt->scan_evt_id) {
  case NRF_BLE_SCAN_EVT_CONNECTING_ERROR: {
    err_code = p_scan_evt->params.connecting_err.err_code;
    APP_ERROR_CHECK(err_code);
  } break;

  case NRF_BLE_SCAN_EVT_CONNECTED: {
    ble_gap_evt_connected_t const *p_connected =
        p_scan_evt->params.connected.p_connected;
    // Scan is automatically stopped by the connection.
    NRF_LOG_INFO("Connecting to target %02x%02x%02x%02x%02x%02x",
        p_connected->peer_addr.addr[0],
        p_connected->peer_addr.addr[1],
        p_connected->peer_addr.addr[2],
        p_connected->peer_addr.addr[3],
        p_connected->peer_addr.addr[4],
        p_connected->peer_addr.addr[5]);
  } break;

  case NRF_BLE_SCAN_EVT_SCAN_TIMEOUT: {
    NRF_LOG_INFO("Scan timed out.");
    scan_start();
  } break;

  default:
    break;
  }
}

/**@brief Function for initializing the scanning and setting the filters.
 */
static void scan_init(void) {
  ret_code_t err_code;
  nrf_ble_scan_init_t init_scan;

  memset(&init_scan, 0, sizeof(init_scan));

  init_scan.connect_if_match = true;
  init_scan.conn_cfg_tag = APP_BLE_CONN_CFG_TAG;

  err_code = nrf_ble_scan_init(&m_scan, &init_scan, scan_evt_handler);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_ble_scan_filter_set(&m_scan, SCAN_NAME_FILTER, &m_target_periph_name);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_ble_scan_filters_enable(&m_scan, NRF_BLE_SCAN_NAME_FILTER, false);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling database discovery events.
 *
 * @details This function is a callback function to handle events from the database discovery module.
 *          Depending on the UUIDs that are discovered, this function forwards the events
 *          to their respective services.
 *
 * @param[in] p_event  Pointer to the database discovery event.
 */
static void db_disc_handler(ble_db_discovery_evt_t *p_evt) {
  ble_nus_c_on_db_disc_evt(&m_ble_nus_c, p_evt);
}

/**@brief Function for handling characters received by the Nordic UART Service (NUS).
 *
 * @details This function takes a list of characters of length data_len and prints the characters out on UART.
 *          If @ref ECHOBACK_BLE_UART_DATA is set, the data is sent back to sender.
 */
static void ble_nus_chars_received_uart_print(uint8_t *p_data, uint16_t data_len) {
  ret_code_t ret_val;

//  NRF_LOG_DEBUG("Receiving data in received_uart_print.");
  NRF_LOG_HEXDUMP_DEBUG(p_data, data_len);

  for (uint32_t i = 0; i < data_len; i++) {
    do {
      ret_val = app_uart_put(p_data[i]);
      if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY)) {
        NRF_LOG_ERROR("app_uart_put failed for index 0x%04x.", i);
        APP_ERROR_CHECK(ret_val);
      }
    } while (ret_val == NRF_ERROR_BUSY);
  }
  if (p_data[data_len - 1] == '\r') {
    while (app_uart_put('\n') == NRF_ERROR_BUSY)
      ;
  }
  if (ECHOBACK_BLE_UART_DATA) {
    // Send data back to the peripheral.
    do {
      ret_val = ble_nus_c_string_send(&m_ble_nus_c, p_data, data_len);
      if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY)) {
        NRF_LOG_ERROR("Failed sending NUS message. Error 0x%x. ", ret_val);
        APP_ERROR_CHECK(ret_val);
      }
    } while (ret_val == NRF_ERROR_BUSY);
  }
}

/**@brief   Function for handling app_uart events.
 *
 * @details This function receives a single character from the app_uart module and appends it to
 *          a string. The string is sent over BLE when the last character received is a
 *          'new line' '\n' (hex 0x0A) or if the string reaches the maximum data length.
 */
void uart_event_handle(app_uart_evt_t *p_event) {
  static uint8_t data_array[BLE_NUS_MAX_DATA_LEN];
  static uint16_t index = 0;
  uint32_t ret_val;

  switch (p_event->evt_type) {
  /**@snippet [Handling data from UART] */
  case APP_UART_DATA_READY:
    UNUSED_VARIABLE(app_uart_get(&data_array[index]));
    index++;

    if ((data_array[index - 1] == '\n') || (index >= (m_ble_nus_max_data_len))) {
      NRF_LOG_DEBUG("Ready to send data over BLE NUS");
      NRF_LOG_HEXDUMP_DEBUG(data_array, index);

      do {
        ret_val = ble_nus_c_string_send(&m_ble_nus_c, data_array, index);
        if ((ret_val != NRF_ERROR_INVALID_STATE) && (ret_val != NRF_ERROR_RESOURCES)) {
          APP_ERROR_CHECK(ret_val);
        }
      } while (ret_val == NRF_ERROR_RESOURCES);

      index = 0;
    }
    break;

  /**@snippet [Handling data from UART] */
  case APP_UART_COMMUNICATION_ERROR:
    NRF_LOG_ERROR("Communication error occurred while handling UART.");
    APP_ERROR_HANDLER(p_event->data.error_communication);
    break;

  case APP_UART_FIFO_ERROR:
    NRF_LOG_ERROR("Error occurred in FIFO module used by UART.");
    APP_ERROR_HANDLER(p_event->data.error_code);
    break;

  default:
    break;
  }
}

void ble_handle_message(char *pdata, int len) {
  unsigned char sdata[20];
  int i, ndata;

  //  Validate good data.

  if (len <= 5 ||
      pdata[0] != 0xfd || pdata[1] != 0xfc ||
      pdata[2] != len - 4) {
    NRF_LOG_INFO("Message reject point 341.");
    goto reject;
  }
  //  Validate check sum

  uint32_t cs = 0;
  for (i = 0; i < (3 + pdata[2]); i++)
    cs += pdata[i];
  cs = 0xff & ((~cs) + 1);
  if (cs != pdata[i]) {
    NRF_LOG_INFO("Message reject checksum, i %d, sb %x, was %x.", i, cs, pdata[i]);
    goto reject;
  }

  user_interface_connect = true;

  //  Analyze message per EPOB0178:

  if (pdata[3] == 0x01 && pdata[2] == 1) {
    NRF_LOG_INFO("Remove valve data cmd.");
    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x81, 0x01, 0x00);
    ndata = 5;
    goto sendreply;
  }

  else if (pdata[3] == 0x02 && pdata[2] == 1) {
    NRF_LOG_INFO("Set BT work mode cmd.");
    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x82, 0x01, 0x00);
    ndata = 5;
    goto sendreply;
  }

  else if (pdata[3] == 0x03 && pdata[2] == 2) {
    NRF_LOG_INFO("Water temp query, scale %d.", pdata[4]);

    temp_scale = pdata[4]; // set the temp_scale, 0 (false)=F, 1 (true) = C
                           // NOTE if temperature scale changes, we need to purge temperature buffer and disable water temperature
                           // changes for at least one second.

    int water_temp = t_avg;
    NRF_LOG_DEBUG("Sending Temp: %d F \r\n", water_temp);
    SETSDATA(0xfd, 0xfc, 0x03, 0x83, 0x00, water_temp); // reply with current water temperature
    ndata = 6;
    goto sendreply;
  }

  else if (pdata[3] == 0x04 && pdata[2] == 3) {
    NRF_LOG_INFO("Water temp cmd, scale %d, degs %d.", pdata[4], pdata[5]);
    temp_scale = pdata[4]; // set the temp_scale, 0 (false)=F, 1 (true) = C
                           // NOTE if temperature scale changes, we need to purge temperature buffer and disable water temperature
                           // changes for at least one second.

    desired_temp = pdata[5];
    //        adjust_temp(pdata[5]);          // Start adjusting temp routine

    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x84, 0x01, 0x00);
    ndata = 5;
    goto sendreply;
  }

  else if (pdata[3] == 0x05 && pdata[2] == 2) {
    NRF_LOG_INFO("Water flow cmd, flow %d.", pdata[4]);
    user_requested_flow(pdata[4]);

    //  Replying success.
    SETSDATA(0xfd, 0xfc, 0x02, 0x85, 0x01, 0x00);
    ndata = 5;
    goto sendreply;
  }

  else {
    NRF_LOG_INFO("Other command received, %x, %x, %x.", pdata[4], pdata[5], pdata[6]);
    //  Not replying.
    goto reject;
  }

sendreply:

  //  Compute checksum.

  cs = 0;
  for (i = 0; i < ndata; i++)
    cs += sdata[i];
  cs = 0xff & ((~cs) + 1);
  sdata[ndata++] = cs;

  //  Send reply.

  uint32_t ret_val = 21; // arbitrary marker number.
  do {
    ret_val = ble_nus_c_string_send(&m_ble_nus_c, sdata, ndata);
    if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY)) {
      NRF_LOG_ERROR("Failed sending NUS message. Error 0x%x. ", ret_val);
      APP_ERROR_CHECK(ret_val);
    }
  } while (ret_val == NRF_ERROR_BUSY);

reject:

  adjust_temp(desired_temp); // Start adjusting temp routine
}

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

/**@brief Callback handling Nordic UART Service (NUS) client events.
 *
 * @details This function is called to notify the application of NUS client events.
 *
 * @param[in]   p_ble_nus_c   NUS client handle. This identifies the NUS client.
 * @param[in]   p_ble_nus_evt Pointer to the NUS client event.
 */

/**@snippet [Handling events from the ble_nus_c module] */
static void ble_nus_c_evt_handler(ble_nus_c_t *p_ble_nus_c, ble_nus_c_evt_t const *p_ble_nus_evt) {
  ret_code_t err_code;

  switch (p_ble_nus_evt->evt_type) {
  case BLE_NUS_C_EVT_DISCOVERY_COMPLETE:
    NRF_LOG_INFO("Discovery complete.");
    err_code = ble_nus_c_handles_assign(p_ble_nus_c, p_ble_nus_evt->conn_handle, &p_ble_nus_evt->handles);
    APP_ERROR_CHECK(err_code);

    err_code = ble_nus_c_tx_notif_enable(p_ble_nus_c);
    APP_ERROR_CHECK(err_code);
    NRF_LOG_INFO("Connected to device with Nordic UART Service.");
    break;

  case BLE_NUS_C_EVT_NUS_TX_EVT:
    ble_nus_chars_received_uart_print(p_ble_nus_evt->p_data, p_ble_nus_evt->data_len);

    /******************************* DFH *************************/
    /******************************* DFH *************************/
    /******************************* DFH *************************/

    NRF_LOG_INFO("NUS TX EVT occurred.");

    ble_handle_message(p_ble_nus_evt->p_data, p_ble_nus_evt->data_len);

    /******************************* DFH *************************/
    /******************************* DFH *************************/
    /******************************* DFH *************************/

    break;

  case BLE_NUS_C_EVT_DISCONNECTED:
    NRF_LOG_INFO("Disconnected.");
    scan_start();
    break;
  }
}
/**@snippet [Handling events from the ble_nus_c module] */

/**
 * @brief Function for handling shutdown events.
 *
 * @param[in]   event       Shutdown type.
 */
static bool shutdown_handler(nrf_pwr_mgmt_evt_t event) {
  ret_code_t err_code;

  err_code = bsp_indication_set(BSP_INDICATE_IDLE);
  APP_ERROR_CHECK(err_code);

  switch (event) {
  case NRF_PWR_MGMT_EVT_PREPARE_WAKEUP:
    // Prepare wakeup buttons.
    err_code = bsp_btn_ble_sleep_mode_prepare();
    APP_ERROR_CHECK(err_code);
    break;

  default:
    break;
  }

  return true;
}

NRF_PWR_MGMT_HANDLER_REGISTER(shutdown_handler, APP_SHUTDOWN_HANDLER_PRIORITY);

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const *p_ble_evt, void *p_context) {
  ret_code_t err_code;
  ble_gap_evt_t const *p_gap_evt = &p_ble_evt->evt.gap_evt;

  switch (p_ble_evt->header.evt_id) {
  case BLE_GAP_EVT_CONNECTED:
    err_code = ble_nus_c_handles_assign(&m_ble_nus_c, p_ble_evt->evt.gap_evt.conn_handle, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_indication_set(BSP_INDICATE_CONNECTED);
    APP_ERROR_CHECK(err_code);

    // start discovery of services. The NUS Client waits for a discovery result
    err_code = ble_db_discovery_start(&m_db_disc, p_ble_evt->evt.gap_evt.conn_handle);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_DISCONNECTED:

    NRF_LOG_INFO("Disconnected. conn_handle: 0x%x, reason: 0x%x",
        p_gap_evt->conn_handle,
        p_gap_evt->params.disconnected.reason);
    break;

  case BLE_GAP_EVT_TIMEOUT:
    if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN) {
      NRF_LOG_INFO("Connection Request timed out.");
    }
    break;

  case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
    // Pairing not supported.
    err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
    // Accepting parameters requested by peer.
    err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
        &p_gap_evt->params.conn_param_update_request.conn_params);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_PHY_UPDATE_REQUEST: {
    NRF_LOG_DEBUG("PHY update request.");
    ble_gap_phys_t const phys =
        {
            .rx_phys = BLE_GAP_PHY_AUTO,
            .tx_phys = BLE_GAP_PHY_AUTO,
        };
    err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
    APP_ERROR_CHECK(err_code);
  } break;

  case BLE_GATTC_EVT_TIMEOUT:
    // Disconnect on GATT Client timeout event.
    NRF_LOG_DEBUG("GATT Client Timeout.");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTS_EVT_TIMEOUT:
    // Disconnect on GATT Server timeout event.
    NRF_LOG_DEBUG("GATT Server Timeout.");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  default:
    break;
  }
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void) {
  ret_code_t err_code;

  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);

  // Configure the BLE stack using the default settings.
  // Fetch the start address of the application RAM.
  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
  APP_ERROR_CHECK(err_code);

  // Enable BLE stack.
  err_code = nrf_sdh_ble_enable(&ram_start);
  APP_ERROR_CHECK(err_code);

  // Register a handler for BLE events.
  NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}

/**@brief Function for handling events from the GATT library. */
void gatt_evt_handler(nrf_ble_gatt_t *p_gatt, nrf_ble_gatt_evt_t const *p_evt) {
  if (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED) {
    NRF_LOG_INFO("ATT MTU exchange completed.");

    m_ble_nus_max_data_len = p_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
    NRF_LOG_INFO("Ble NUS max data length set to 0x%X(%d)", m_ble_nus_max_data_len, m_ble_nus_max_data_len);
  }
}

/**@brief Function for initializing the GATT library. */
void gatt_init(void) {
  ret_code_t err_code;

  err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_ble_gatt_att_mtu_central_set(&m_gatt, NRF_SDH_BLE_GATT_MAX_MTU_SIZE);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling events from the BSP module.
 *
 * @param[in] event  Event generated by button press.
 */
void bsp_event_handler(bsp_event_t event) {
  ret_code_t err_code;

  switch (event) {
  case BSP_EVENT_SLEEP:
    nrf_pwr_mgmt_shutdown(NRF_PWR_MGMT_SHUTDOWN_GOTO_SYSOFF);
    break;

  case BSP_EVENT_DISCONNECT:
    err_code = sd_ble_gap_disconnect(m_ble_nus_c.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    if (err_code != NRF_ERROR_INVALID_STATE) {
      APP_ERROR_CHECK(err_code);
    }
    break;

  default:
    break;
  }
}

/**@brief Function for initializing the UART. */
static void uart_init(void) {
  ret_code_t err_code;

  app_uart_comm_params_t const comm_params =
      {
          .rx_pin_no = RX_PIN_NUMBER,
          .tx_pin_no = TX_PIN_NUMBER,
          .rts_pin_no = RTS_PIN_NUMBER,
          .cts_pin_no = CTS_PIN_NUMBER,
          .flow_control = APP_UART_FLOW_CONTROL_DISABLED,
          .use_parity = false,
          .baud_rate = UART_BAUDRATE_BAUDRATE_Baud115200};

  APP_UART_FIFO_INIT(&comm_params,
      UART_RX_BUF_SIZE,
      UART_TX_BUF_SIZE,
      uart_event_handle,
      APP_IRQ_PRIORITY_LOWEST,
      err_code);

  APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the Nordic UART Service (NUS) client. */
static void nus_c_init(void) {
  ret_code_t err_code;
  ble_nus_c_init_t init;

  init.evt_handler = ble_nus_c_evt_handler;

  err_code = ble_nus_c_init(&m_ble_nus_c, &init);
  APP_ERROR_CHECK(err_code);
}

/* Function for initializing timer, pins, buttons and leds. */
static void pins_buttons_leds_init(void) {
    ret_code_t err_code;
    bsp_event_t startup_event;

    err_code = bsp_init(BSP_INIT_LEDS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);

    // Set up our system timer for general operational use
//    const nrf_drv_timer_t sys_timer = NRF_DRV_TIMER_INSTANCE(1); // use timer 1 for non-softdevice timers was 0
                                                                 //    APP_TIMER_DEF(sys_timer);                 
                                                                 // Creates timer id for our program.
  // ***************** Set up system I/Os **********************
  // ***********************************************************
  // NOTE: D_TEMP input, AKA ONEWIRE is is bidirectional and is controlled in temperature routine
    nrf_gpio_cfg_output(ONEWIRE_PIN);
    nrf_gpio_pin_set(ONEWIRE_PIN);

    // Set STEP_A and STEP_B pins as standard outputs and set the outputs low
    nrf_gpio_cfg_output(STEP_A); // Set pins as outputs
    nrf_gpio_cfg_output(STEP_B);
    nrf_gpio_pin_clear(STEP_A); // Set pins low.
    nrf_gpio_pin_clear(STEP_B);

    // Set up buzzer pin as standard output.
    nrf_gpio_cfg_output(AUDIO_OUT);
    //    nrf_gpio_pin_clear(AUDIO_OUT);         // Set pin low.
    nrf_gpio_pin_set(AUDIO_OUT); // Set pin high.

    // Set up LEDs as standard outputs.
    nrf_gpio_cfg_output(LED_1);
    nrf_gpio_pin_clear(LED_1); // Set pin low. Turns on LED.
    nrf_gpio_cfg_output(LED_2);
    nrf_gpio_pin_clear(LED_2); // Set pin low. Turns on LED.

    // Set up a_temp_out as standard output
    nrf_gpio_cfg_output(A_TEMP_OUT);
    nrf_gpio_pin_clear(A_TEMP_OUT); // Set pin low.

    // Set up Sys_V as standard input w/ no pullup.
    nrf_gpio_cfg_input(SYS_V, NRF_GPIO_PIN_NOPULL);

    // Set up A_temp_in as standard input w/ no pullup.
    nrf_gpio_cfg_input(A_TEMP_IN, NRF_GPIO_PIN_NOPULL);

    // Set up trip_test as a standard output with pull down;
    NRF_GPIO->PIN_CNF[TRIP_TEST] = (NRF_GPIO_PIN_NOSENSE) | (NRF_GPIO_PIN_S0S1) | (NRF_GPIO_PIN_PULLDOWN) | (NRF_GPIO_PIN_INPUT_DISCONNECT) | (NRF_GPIO_PIN_DIR_OUTPUT);

    // Set up temp_override pin as open-collector high drive output
    NRF_GPIO->PIN_CNF[TEMP_OVERRIDE] = (NRF_GPIO_PIN_NOSENSE) | (NRF_GPIO_PIN_H0D1) | (NRF_GPIO_PIN_NOPULL) | (NRF_GPIO_PIN_INPUT_DISCONNECT) | (NRF_GPIO_PIN_DIR_OUTPUT);
    nrf_gpio_pin_clear(TEMP_OVERRIDE); // Set pin high to disable stepper.

    // Set up manual override and manual reset pins as sensing inputs w/pullups
    nrf_gpio_cfg_sense_input(MAN_OVERRIDE, NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_HIGH); // SENSE_HIGH OR LOW?
    nrf_gpio_cfg_sense_input(MAN_RESET, NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_HIGH);    // SENSE_HIGH OR LOW?
    nrf_gpio_cfg_sense_input(VALVE_OFF, NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_HIGH);    // SENSE_HIGH OR LOW?
}
/*
// GPIOTE is used for D_TEMP_IRQ input to know when a temperature conversion is completed
void in_pin_handler(nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action) {
  get_water_temp();
}

/**
 * @brief Function for configuring: PIN_IN pin for input, 
 * and configures GPIOTE to give an interrupt on pin change. 

#define PIN_IN D_TEMP_IRQ

void gpio_init(void) {
  ret_code_t err_code;
  err_code = nrf_drv_gpiote_init();
  APP_ERROR_CHECK(err_code);

  nrf_drv_gpiote_in_config_t in_config = GPIOTE_CONFIG_IN_SENSE_LOTOHI(true);
  in_config.pull = NRF_GPIO_PIN_NOPULL;

  err_code = nrf_drv_gpiote_in_init(PIN_IN, &in_config, in_pin_handler);
  APP_ERROR_CHECK(err_code);

//  d_temp_irq_enable();
}
//  Functions to enable and disable interrupts on d_temp_irq
int d_temp_irq_enable(void) {
  nrf_drv_gpiote_in_event_enable(PIN_IN, true);
//  temp_probe_irq = true;
}

uint32_t d_temp_irq_disable(void) {
  nrf_drv_gpiote_in_event_disable(PIN_IN);
//  temp_probe_irq = false;
}

  // **********************  Setup PWM peripheral  **************************
    #define OUTPUT_PIN AUDIO_OUT
    static nrf_drv_pwm_t m_pwm0 = NRF_DRV_PWM_INSTANCE(0);

    // Declare variables holding PWM sequence values. In this example only one channel is used 
    nrf_pwm_values_individual_t seq_values[] = {0x8000, 0, 0x8000, 0};       // 0x8000 is equal to 200ms

    nrf_pwm_sequence_t const seq =
    {
        .values.p_individual = seq_values,
        .length          = NRF_PWM_VALUES_LENGTH(seq_values),
        .repeats         = 0,
        .end_delay       = 0
    };
*/
/******************** END OF PIN CONFIGURATIONS *************************/
/*****************************************************************************************
                Set up timers for our general use
******************************************************************************************/
APP_TIMER_DEF(read_temperature_timer);
static void read_temperature_timer_handler(void * p_context)
{
  nrf_gpio_pin_toggle(LED_2); // Flash the led so we know something's happening
  get_water_temp();
}

APP_TIMER_DEF(stepper_delay_timer);
void stepper_delay_timer_handler(nrf_timer_event_t event_type, void *p_context) 
{
   stepper_delay();
}

static void app_timers_init(void) 
{
  // Create timers
  ret_code_t err_code;
  err_code = app_timer_create(&read_temperature_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              read_temperature_timer_handler);

  err_code = app_timer_create(&stepper_delay_timer,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              stepper_delay_timer_handler);
}

// Start the Temperature Conversion timer
void start_temperature_conversion_timer(convert_time)
{
  uint32_t time_in_ticks;
//  time_in_ticks = nrf_drv_timer_ms_to_ticks(&read_temperature_timer, convert_time);
  time_in_ticks = convert_time;

  uint32_t err_code;
  err_code = app_timer_start(read_temperature_timer, APP_TIMER_TICKS(time_in_ticks), NULL);
  APP_ERROR_CHECK(err_code);
}

void stepper_delay_start(void) 
{
  ret_code_t err_code;
  err_code = app_timer_start(stepper_delay_timer, APP_TIMER_TICKS(STEPPER_MOTOR_RATE), NULL); // mS between steps
  APP_ERROR_CHECK(err_code);
}

void stepper_delay_stop(void) 
{
  ret_code_t err_code;
  err_code = app_timer_stop(stepper_delay_timer);
  APP_ERROR_CHECK(err_code);
}
/******************** END OF Timers CONFIGURATIONS *************************/

/**@brief Function for initializing the nrf log module. */
static void log_init(void) {
  ret_code_t err_code = NRF_LOG_INIT(NULL);
  APP_ERROR_CHECK(err_code);

  NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/**@brief Function for initializing power management.
 */
static void power_management_init(void) {
  ret_code_t err_code;
  err_code = nrf_pwr_mgmt_init();
  APP_ERROR_CHECK(err_code);
}

/** @brief Function for initializing the database discovery module. */
static void db_discovery_init(void) {
  ret_code_t err_code = ble_db_discovery_init(db_disc_handler);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling the idle state (main loop).
 *
 * @details Handles any pending log operations, then sleeps until the next event occurs.
 */
static void idle_state_handle(void) {
  if (NRF_LOG_PROCESS() == false) {
    nrf_pwr_mgmt_run();
  }
}

int main(void) {
  // Initialize.
  log_init();
  app_timer_init();         // Initialize timer lib
  app_timers_init();        // Setup the system timers

  uart_init();
  pins_buttons_leds_init();    // Eric - modified
//  gpio_init();                 // Eric - NO LONGER USED
//  pwm_init();                  // Eric added

//  saadc_init();                // Eric - added
//  saadc_sampling_event_init(); // Eric - added
  perfect_shower_init();       // Eric

  db_discovery_init();
  power_management_init();
  ble_stack_init();
  gatt_init();
  nus_c_init();
  scan_init();

  //    perfect_shower_main();                        // Eric Perfect Shower main routines

  // Start execution.
  printf("BLE UART central example started.\r\n");
  NRF_LOG_INFO("BLE UART central example started.");
  scan_start();

  // Enter main loop.
  for (;;) {
    idle_state_handle();
  }
}

// *************************************************************************************
/* **************** End of Nordic ble_app_uart_c Example Code **************************
// *************************************************************************************/

// *************************************************************************************
/* **************** Start of Perfect Shower Main Routine **************************
// *************************************************************************************/
/********* Move to top of main when done *******************/

    // This routine is entered into via a timer expiration. Used to perfrom housekeeping
int perfect_shower_main() {
  printf("Perfect Shower Main started.\r\n");
  NRF_LOG_INFO("Perfect Shower Main started.");

//  d_temp_irq_disable(); // disable irq's from temperature probe until we are ready for them

  nrf_gpio_pin_toggle(LED_2); // Flash the led so we know something's happening

  //    int current_temp = get_water_temp();
  // t_avg is the average of last 5 temperature readings

//  rotate_valve(int32_t 5, char heat) 

  // ****************** Enable the stepper motor - Reduces motor heating during testing *****************
  nrf_gpio_pin_set(TEMP_OVERRIDE);
  nrf_delay_ms(1); // Give motor driver some time.

  // ************************** End *********************************************************************

  if (current_temp < 85) { // current_temp is either F or C
    rotate_valve(20, "heat");
  } else {
    rotate_valve(20, "cool");
  }
  nrf_delay_ms(1000); // give us a delay to monitor the stepper

  // ****************** Disable the stepper motor - Reduces motor heating during testing *****************
  nrf_gpio_pin_clear(TEMP_OVERRIDE);
  nrf_delay_ms(1); // Give motor driver some time.
  // ************************** End *********************************************************************
}

void adjust_temp(int requested_temp) {
  flashblock.desired_temp = requested_temp;

  // This is where we adjust the valve to get the actual water temperature to match the desired temperature
  if (requested_temp != t_avg) { // determine how far off the temperature is
    int delta_t = abs(requested_temp - t_avg);

    if (delta_t <= 2) {
      // step valve by 1 step (determine which direction)

    } else if (delta_t > 2 && delta_t <= 5) {
      // Step valve by 2 steps (determine which direction)

    } else {
      // step valve by 3 steps (determine which direction)
    }
  }

} // End of function

// *************** We don't control flow rate yet so we'll ignore for now **************
void user_requested_flow(int flow) {
  flashblock.desired_flow = flow;
  //  We'll use this function to turn on the valve
}
//

void write_flashblock() {
  // bunch of NRF stuff.
}

void read_flashblock() {
  // bunch of NRF stuff.
  temp_scale = false; // C or F. True = C, False = F. This is set by the user via the display
}
// *************************************************************************************
//                        End - Perfect Shower Main Code
// *************************************************************************************
// *************************************************************************************
//       Nordic DS18B20 Temperature sensor interface (FINISHED) (Eric's version)
// *************************************************************************************

enum {
  BSPACM_ONEWIRE_CMD_READ_ROM = 0x33,          // Read 64-bit ROM code without using search procedure
  BSPACM_ONEWIRE_CMD_SKIP_ROM = 0xCC,          // Skip ROM sends the following command to all bus devices
  BSPACM_ONEWIRE_CMD_READ_POWER_SUPPLY = 0xB4, // Tw]ƬOHͨѹq٬O~ѹq
  BSPACM_ONEWIRE_CMD_RECALL_EE = 0xB8,         // Store data from EEPROM into RAM
  BSPACM_ONEWIRE_CMD_READ_SCRATCHPAD = 0xBE,   // Read the RAM area.
  BSPACM_ONEWIRE_CMD_CONVERT_T = 0x44,         // Initiate a temperature conversion.
  BSPACM_ONEWIRE_CMD_WRITE_SCRATCHPAD = 0x4E,  // Write to scratchpad memory command
};

enum {
  OWT_RSTL_us = 480,                          // Minimum time to hold bus low to ensure reset
  OWT_RSTH_us = 480,                          // Time to wait for presence detection after reset to quiesce
  OWT_PDHIGH_us = 60,                         // Delay before presence pulse is known to be valid (15us..60us)
  OWT_LOW0_us = 60,                           // Minimum time to hold bus low when writing a zero
  OWT_LOW1_us = 5,                            // Eric was 1   	// Minimum time to hold bus low when writing a one
  OWT_REC_us = 5,                             // Eric was 1	// Recovery delay between write/read transaction cycles
  OWT_INT_us = 5,                             // Eric was 1	// Time to hold bus low when initiating a read slot
  OWT_RDV_us = 25 - OWT_INT_us,               // Eric was 15	// Point at which read value should be sampled
  OWT_SLOT_us = 80,                           // Eric was 60	// Minimum duration of a read or write slot
  BSPACM_ONEWIRE_CMD_WRITE_Th_Reg = 0x34,     // 34 = two's complement of 52C - the high temperature threshold
  BSPACM_ONEWIRE_CMD_WRITE_Tl_Reg = 0x00,     // 00 = two's complement of 0C - the Low temperature threshold
  BSPACM_ONEWIRE_CMD_WRITE_CONFIG_Reg = 0x3F, // Set up config reg for 10 bit conversion.
  TEMP_CONVERT_TIME_9BIT = 90,                // Time in ms to complete temperature conversion.
  TEMP_CONVERT_TIME_10BIT = 190,              // s/b 190 - Time in ms to complete temperature conversion.
  TEMP_CONVERT_TIME_11BIT = 380,              // Time in ms to complete temperature conversion.
  TEMP_CONVERT_TIME_12BIT = 755,              // Time in ms to complete temperature conversion.
};

/*
volatile bool cc0_timeout;

void TIMER0_IRQHandler ()                  // Eric - Set up timer 1 for this because of soft-device?
{
  if (NRF_TIMER0->EVENTS_COMPARE[0]) {     // Eric - Set up timer 1 for this because of soft-device?
    cc0_timeout = true;
    NRF_TIMER0->EVENTS_COMPARE[0] = 0;
  }
}

void delay_us (uint32_t count_us)
{

  /* Don't muck with timers if the duration is so short we might miss
   * the compare value wrap. 

  if (2 >= count_us) {
    nrf_delay_us(count_us);
    return;
  }
  cc0_timeout = false;
  NRF_TIMER0->TASKS_CAPTURE[0] = 1;
  NRF_TIMER0->EVENTS_COMPARE[0] = 0;
  NRF_TIMER0->INTENSET = TIMER_INTENSET_COMPARE0_Msk;
  NRF_TIMER0->CC[0] += count_us;
  while (! cc0_timeout) {
    __WFE();
  }
  NRF_TIMER0->INTENCLR = TIMER_INTENCLR_COMPARE0_Msk;
  return;
}
*/

#define ONEWIRE_BIT (1ULL << ONEWIRE_PIN)

static void dirset(void) // Set's pin as output.
{
  NRF_GPIO->DIRSET = ONEWIRE_BIT;
}

void dirclr(void) // Set's pin as input.
{
  NRF_GPIO->PIN_CNF[ONEWIRE_PIN] = 0 | (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos) | (GPIO_PIN_CNF_DRIVE_S0S1 << GPIO_PIN_CNF_DRIVE_Pos) | (GPIO_PIN_CNF_PULL_Disabled << GPIO_PIN_CNF_PULL_Pos) | (GPIO_PIN_CNF_INPUT_Connect << GPIO_PIN_CNF_INPUT_Pos) | (GPIO_PIN_CNF_DIR_Input << GPIO_PIN_CNF_DIR_Pos);
}

/* ************************** Set up the system at power up ************************
************************************************************************************/

void perfect_shower_init(void) {

/*************** READ THE FLASH SETTINGS **************************
*
* call "read flash function" here to get system settings.
*
*******************************************************************/

/***************  First get the temperature sensor working  *********/
    int t_probe_present = false;
    if (0 == iBSPACMonewireReset_ni()) {
      printf("ERR: No DS18B20 present on P0.%u\n", ONEWIRE_PIN);
      NRF_LOG_INFO("ERR: No DS18B20 present on P0.%u\n", ONEWIRE_PIN);
      goto exit_probe_init;
    } else {                                                           
      t_probe_present = true;                                          // Temp device responded.
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);         // Send the SKIP ROM command.
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_SCRATCHPAD); // Write command
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_Th_Reg);     // Set the high temperature threshold
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_Tl_Reg);     // Set the low temperature threshold
      vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_WRITE_CONFIG_Reg); // Write 3F to configuration register for 10 bit res
    }
            // start the temperature probe to perform a temperature conversion
    perform_temp_conversion();
    dirclr();                                                          // make sure pin is set as input
            // Set irq to wait for the conversion to complete. Onewire pin will go high. Takes 187ms for 10 bit accuracy
//    d_temp_irq_enable();
      // Start conversion timer
    start_temperature_conversion_timer(TEMP_CONVERT_TIME_10BIT);

            // Start clock for accurate frequencies
//    NRF_CLOCK->TASKS_HFCLKSTART = 1; 
            // Wait for clock to start
//    while(NRF_CLOCK->EVENTS_HFCLKSTARTED == 0);

//    nrf_drv_pwm_simple_playback(&m_pwm0, &seq, 1, NRF_DRV_PWM_FLAG_LOOP); // Start PWM output

            // Turn on the PWM for the buzzer
//    nrf_drv_pwm_simple_playback(&m_pwm0, &seq, 1, NRF_DRV_PWM_FLAG_LOOP);

            // Flush the hot water
    purge_valve();

// void perfect_shower_install (void) {
    if (install_mode && !initialization_done) {  // This code only runs during Installation.

  /*************************************************************************
   *                  VALVE ROTATION DIRECTION
   *  This code figures out which direction of rotation makes the water
   *  hotter and colder.  The valve could be plumbed backwards and different
   *  manufacturers have different designs or the stepper motor could be
   *  installed in the WaterGeni upside down.
   *  NOTE: This code is only run during initial installation or after a
   *  reset.  The "bool clockwiseHeat = true/false" is stored in flash for 
   *  subsequent reference during power up so the valve knows which way to turn.
   *
   *  The valve is first powered up after installation, by manually pulling out 
   *  the handle and rotating it.  The valve will be in "Manual" mode because
   *  the handle is pulled out to engage the handle and water will be flowing.  
   *  If the handle is not pulled out the handle will not be engaged and the 
   *  valve will not turn on.  Turn handle approximately 50% then push in to 
   *  engage automatic mode.
   *************************************************************************/
    printf("Perfect Shower Install Started.\r\n");
    NRF_LOG_INFO("Perfect Shower Install Started.\n");
    nrf_delay_ms(1500);                          // time to get 5 temperature readings for t_avg 

        // turn on valve approx 50%.  Since the valve has a 95 degree rotation from 
        // off to full hot and there 200 steps per rotation on the stepper, are there are
        // 1.8 degrees per step.  That means that there are approximately 53 steps from
        // off to full hot.  50% is about 26 steps.
    stabilize_temperature();
    int32_t curr_temp_1 = current_temp;       // use current temperature not average temp
        // Measure system voltage (Sys_V) pin. Will be approx 1.0V when valve ON and water flowing.
    int32_t sys_v1 = get_sys_v();             // read the system voltage to determine valve OFF condition.
        // rotate valve 8 steps to cool water and recheck temp. 
        // (should be colder water if valve installed correctly.)
    rotate_valve(8, "cool"); 
    stabilize_temperature();
    int32_t curr_temp_2 = current_temp;       // use current temperature not average temp
    if (curr_temp_1 > curr_temp_2){
      clockwise_heat = true;
    }else {
      clockwise_heat = false;                 // valve installed backwards or stepper backwards
    }

        // Determine which direction to rotate valve to turn off water
        // Measure system voltage (Sys_V) pin again. If water flow is decreased enough Sys_V will be lower.
    int32_t sys_v2 = get_sys_v();             // read the system voltage
    while (sys_v2 == sys_v1) {
      rotate_valve(10, "cool");             // // If sys_v hasn't changed then turn 10 more steps cooler.
      nrf_delay_ms(500);                          // wait a short time.
      int32_t sys_v2 = get_sys_v();           // read the system voltage
    }
    
    if(sys_v1 > sys_v2){
      valve_off_dir[4] = "cool";
    } else {
      valve_off_dir[4] = "heat";
    }

        // Rotate valve to full hot and measure the maximum HOT temperature of the water source
    if (valve_off_dir[4] == "cool"){
      rotate_valve(50, "heat");               // turn valve all the way to full hot. Valve stops at full ON.
      stabilize_temperature();
      users_hot_water_temp = t_avg;             // read the water temperature
    } else {

// change the code below to detect the valve OFF input indicating that we are close to shutting off the valve.

        do {                                    // turn valve towards OFF until water flow slows to about half
          rotate_valve(1, "heat");              // turn valve all the way to full hot.
          nrf_delay_ms(500);                        // wait a short time for system to settle
          sys_v2 = get_sys_v();                 // read the system voltage
        } while (sys_v2 > 0.5);                 // the voltage is approx 1.0V @ 18Vgen, so about 0.5V @ 9Vgen.
        stabilize_temperature();
        users_hot_water_temp = t_avg;           // read the water temperature
      }

//  Save initialization settings to flash here- Call Flash Routine here 
    install_mode = false;
    initialization_done = true;
//  clockwise_heat;
//  valve_off_dir[];
//  users_hot_water_temp;

// call flash write routine //
    }

    handle_error:
    exit_probe_init:
    return 99; // Send some error code.

} // End of initialization
/**************************** End of initialization ****************************/

/***************  Purge the hot water supply  **********************/
void purge_valve(void)
{
/*
    int32_t curr_temp_1;

    if(valve_off_dir[4] == "cool") {
      rotate_valve(55, "heat");               // turn valve all the way to full hot. Valve stops at full ON.
    } else {
      do {
        rotate_valve(1, "heat");                // turn valve all the way till valve_off input is detected
      } while (VALVE_OFF);
    }
    curr_temp_1 = current_temp;
    while (curr_temp_1 < (desired_temp - 5)){
      curr_temp_1 = current_temp;
    }
    adjust_temp(desired_temp);
*/
}

/**************  Wait for water temp to stabilize ******************/
void stabilize_temperature(void)
{
/*
  int32_t count_to_3 = 0;
  int32_t temp_1, temp_2;
    // wait for the water temperature to stabilize for 15 seconds
    do{
       do{
        int32_t temp_1 = t_avg;                  // get the average temperature
        nrf_delay_ms(5000);                          // wait 5 seconds then take another reading
        temp_2 = t_avg;
       } while (temp_2 != temp_1);
       count_to_3 ++;
    } while (count_to_3 < 4);  
*/
}

int get_water_temp(void) {
//  ***** This routine is ONLY entered via the read_temperture_timer timeout.
//  ***** All other system temperature functions should use global variables such as: 
//  ***** desired_temp, user_temp, raw_temp, current_temp and t_ave, some of which are set herein.
//  ***** This routine runs constantly in the background at the speed of the Dallas sensor.

//  NRF_LOG_INFO("Entered get_water_temp routine");
  int32_t raw_temp;                                      // raw DS18B20 temperature sensor data

//  d_temp_irq_disable();                                  // disable the interrupt so we don't generate 
                                                         // new interrupts while reading the sensor
  // Temperature ready - Read it and convert it.
  t_raw = read_temp_probe(&t_raw);
  raw_temp = t_raw;
  raw_temp = raw_temp & ~3;                              // ignore last two bits (for 10 bit precicion)
  t_dCel = (float)raw_temp / 16.0;
  t_dFahr = t_dCel * 1.8 + 32.0;

  if (temp_scale) {                                        // for Celsius...
    int intpart = (int)t_dCel;
    float decimal_part = t_dCel - intpart;                 // just the decimal part
    if ((decimal_part > 0.5) && (decimal_part < 0.9)) {
      current_temp = (intpart << 1) | 0x0001;              // set bit 0 to a 1 for 0.5C
    } else {
      current_temp = (intpart << 1);
    }
  } else {                                                 // for farenheit... Round up
    int intpart = (int)t_dFahr;
    float decimal_part = t_dFahr - intpart;                // just the decimal part
    if ((decimal_part > 0.5) && (decimal_part < 0.9)) {
      current_temp = t_dFahr + 1;                          // round up
    } else {
      current_temp = t_dFahr;                              // We only send intergers for F
    }
  }

//  printf("Got %x t_raw, %d C, %d F \r\n", t_raw, t_dCel, t_dFahr);
//  NRF_LOG_INFO("Got %x t_raw, %d C, %d F \n", t_raw, t_dCel, t_dFahr);

      // For some reason, the temperature probe sometimes returns whacked out high readings. 
      // Check for bad temperature readings and limit the temperature to a max difference of 
      // 10 degrees so we don't screw up our avg temp calc or our system response time.
  if ((!temp_scale && (current_temp >= t_avg +10)) || (temp_scale && (current_temp >= t_avg + 5)))
    {
      if(!temp_scale)
        { 
        current_temp = (t_avg + 10);
        }
      else
        {
         current_temp = (t_avg + 4);
        }
    }
                                                           
  tdata[t_count] = current_temp;                           // place in storage and calculate new average  
  if (t_count > 2) {                                       // 3 readings is about a half second
    t_count = 0;
  } else {
    t_count++;
  }
      // Average the 3 readings.  Three readings chosen for a fast response time.
  t_avg = (tdata[0] + tdata[1] + tdata[2]) / 3;

// Debug code:
  if ((t_avg < 68) || (t_avg > 98)){
    NRF_LOG_DEBUG("Temp out of range: \r\n %d %d %d %d", t_avg, tdata[0], tdata[1], tdata[2]);
  }

bad_value:
      // Now, start a new temperature conversion cycle
  perform_temp_conversion();
  start_temperature_conversion_timer(TEMP_CONVERT_TIME_10BIT);

  return t_avg;
}

//  ************* Dallas Temp Sensor Reset  ********************
int iBSPACMonewireReset_ni(void) {
  int present = 0;

  dirset();
  NRF_GPIO->OUTSET = ONEWIRE_BIT; // Set pin high
  nrf_delay_us(480);              // Start by holding bus high for 480uS

  // Hold bus low for OWT_RESET_us
  NRF_GPIO->OUTCLR = ONEWIRE_BIT; // Set pin low
  nrf_delay_us(600);              // Hold bus low for at least 480uS

  // Release bus and switch to input until presence pulse should be
  // visible.
  dirclr();          // Set pin as input
  nrf_delay_us(180); //  Eric - Wait at least 60uS before reading the bus.
                     // Read the bus in the middle of the cycle (60 +120).

  // Record presence if bus is low (DS182x is holding it there)
  present = !!(NRF_GPIO->IN & ONEWIRE_BIT); // Read the pin and invert. Present = 1 means the device present.

  // Wait for reset cycle to complete
  nrf_delay_us(300); //  Eric wait the full 480uS to make sure the DS18B20 is done.

  return present;
}

void vBSPACMonewireWriteByte_ni(int byte) {
  int bp;
  dirclr();        // Make sure bus is high to begin with. Setting pin as input so pullup sets to 1
  nrf_delay_us(5); // All we need is 1uS or so.

  for (bp = 0; bp < 8; ++bp) {
    NRF_GPIO->OUTCLR = ONEWIRE_BIT; // Set pin output low
    dirset();                       // now, set pin as output to drive bus low.
    if (byte & 0x01) {
      nrf_delay_us(8);  // Write a 1
      dirclr();         // Release the bus by setting pin as input so pullup sets to 1
      nrf_delay_us(60); // hold till end of write 1 cycle.
    } else {            // Write a 0
      nrf_delay_us(60); // Hold bus low to write 0
      dirclr();         // Release the bus by setting pin as input so pullup sets to 1
    }
    byte >>= 1;
    nrf_delay_us(5); // Min recovery time > 2uS before next bit.
  }
}

int iBSPACMonewireReadBit_ni(void) {
  int rv;
  dirclr();                          // Make sure bus is high to begin with.
                                     // Set pin as input so pullup sets to 1
  nrf_delay_us(5);                   // All we need is a minimum of 1uS
  NRF_GPIO->OUTCLR = ONEWIRE_BIT;    // Set pin output low
  dirset();                          // set pin as output to drive low.
  nrf_delay_us(5);                   // keep low for a min of 1uS
  dirclr();                          // release bus to read bit.
  nrf_delay_us(8);                   // wait till ready for read
  rv = (NRF_GPIO->IN & ONEWIRE_BIT); // read the value
  nrf_delay_us(45);                  // wait for device to release bus

  return rv;
}

int iBSPACMonewireReadByte_ni(void) {
  int byte = 0;
  int bit = 1;

  do {
    if (iBSPACMonewireReadBit_ni()) {
      byte |= bit;
    }
    bit <<= 1;
  } while (0x100 != bit);
  return byte;
}

int perform_temp_conversion(void) {
  if (!iBSPACMonewireReset_ni()) {
    return -1;
  }
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_CONVERT_T);
  dirclr();   // return ONEWIRE_PIN to input
  //  Go and wait for timer signaling that Dallas sensor competed conversion   
  return 0;
}

int read_temp_probe(int *temp_xCel) {
  int t;

  if (!iBSPACMonewireReset_ni()) {
    return -1;
  }
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
  nrf_delay_us(25); // delay for testing purposes
  vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_READ_SCRATCHPAD);
  nrf_delay_us(25); // delay for testing purposes
  t = iBSPACMonewireReadByte_ni();
  nrf_delay_us(25); // delay for testing purposes
  t |= (iBSPACMonewireReadByte_ni() << 8);

  return t;
}

/******************************************************************************/
//            End of Nordic DS18B20 Temperature sensor module
// ****************************************************************************

/*********************************************************************
*                   Stepper Motor Control (FINISHED)
**********************************************************************/

// Stepper Motor Step sequiences for STEP_A & STEP_B GPIOs
// #define STEP_CLOCKWISE[4] {00, 10, 11, 01}             // 0,2,3,1
// #define STEP_COUNTERCLOCKWISE[4] {00, 01, 11, 10}      // 0,1,3,2

uint32_t step_pin, stepper_A, stepper_B;        // temporary value of output pin STEP_A & STEP_B
char heat_cool[4];
int32_t num_steps;

/********************  Begin of Stepper Motor Delay Code version  ********************/
/*
void rotate_valve(int32_t num_of_steps, char heat_cool){

      // determine which way we need to step in order to heat or cool
    if (clockwise_heat && (heat_cool == "heat")){
      step_forward(num_of_steps);            

    } else if (clockwise_heat && (heat_cool == "cool")){
      step_backward(num_of_steps);

    } else if (!clockwise_heat && (heat_cool == "heat")){
      step_backward(num_of_steps);

    } else if (!clockwise_heat && (heat_cool == "cool")){
      step_forward(num_of_steps);
    }
}

      // Step Forward Direction
 void step_forward(uint32_t num_steps) {        
    stepper_A = nrf_gpio_pin_out_read(STEP_A);    // Read the current state of STEP_A and STEP_B pin levels
    stepper_B = nrf_gpio_pin_out_read(STEP_B);

    if (stepper_A && stepper_B || !stepper_A && !stepper_B) { // Determine which drive pin (STEP_A or STEP_B) needs to be
      step_pin = STEP_B;                                      // changed first in order to move in the forward direction.
    } else {
      step_pin = STEP_A;
    }
          // Now Step the requested number of steps
    do {
      nrf_gpio_pin_toggle(step_pin); // First step in the forward direction
      if (step_pin == STEP_B) {
        step_pin = STEP_A;
      } else {
        step_pin = STEP_B;
      }
      nrf_delay_ms(STEPPER_MOTOR_RATE); // X mS delay for motor. Change this to a timer
      num_steps--;
    } while (num_steps != 0);
}

      // Step Backward direction
 void step_backward(uint32_t num_steps) {     // eliminate passing parameters so we can use timers
    stepper_A = nrf_gpio_pin_out_read(STEP_A); // Read the current state of STEP_A and STEP_B pin levels
    stepper_B = nrf_gpio_pin_out_read(STEP_B);

    if (stepper_A && stepper_B || !stepper_A && !stepper_B) { // Determine which drive pin (STEP_A or STEP_B) needs to be
      step_pin = STEP_A;                                      // changed first in order to move in the backward direction.
    } else {
      step_pin = STEP_B;
    }
      // Now Step the requested number of steps
    do {
      nrf_gpio_pin_toggle(step_pin); // First step in the backward direction
      if (step_pin == STEP_B) {
        step_pin = STEP_A;
      } else {
        step_pin = STEP_B;
      }
      nrf_delay_ms(STEPPER_MOTOR_RATE); // 50mS delay. Change this to timer ticks
      num_steps--;
    } while (num_steps != 0);
}
/*
void stepper_delay(void){
  if (step_forward_timer_started == true){
    step_forward();
  } else if (step_backward_timer_started == true) {
    step_backward();
  }
}
*/
/********************  End of Stepper Motor Delay Code version  ********************/
/******************  Begin of Stepper Motor Timer Code verison  ********************/

void rotate_valve(int32_t num_of_steps, char heat_cool){
    int32_t num_steps = num_of_steps;  

      // determine which way we need to step in order to heat or cool
    if (clockwise_heat && (heat_cool == "heat")){
//      step_forward(num_of_steps);               // eliminate passing parameters so we can use timers
      step_forward();

    } else if (clockwise_heat && (heat_cool == "cool")){
//      step_backward(num_of_steps);
      step_backward();

    } else if (!clockwise_heat && (heat_cool == "heat")){
//      step_backward(num_of_steps);
      step_backward();

    } else if (!clockwise_heat && (heat_cool == "cool")){
//      step_forward(num_of_steps);
      step_forward();
    }
}

      // Step Forward Direction
// void step_forward(uint32_t num_steps) {        // eliminate passing parameters so we can use timers
void step_forward(void) {

    if (step_forward_timer_started == true){
//      step_forward_timer_started = false;
      goto next_for_step;
    }
    stepper_A = nrf_gpio_pin_out_read(STEP_A);    // Read the current state of STEP_A and STEP_B pin levels
    stepper_B = nrf_gpio_pin_out_read(STEP_B);

    if (stepper_A && stepper_B || !stepper_A && !stepper_B) { // Determine which drive pin (STEP_A or STEP_B) needs to be
      step_pin = STEP_B;                                      // changed first in order to move in the forward direction.
    } else {
      step_pin = STEP_A;
    }
          // Now Step the requested number of steps
    do {
      nrf_gpio_pin_toggle(step_pin); // First step in the forward direction
      if (step_pin == STEP_B) {
        step_pin = STEP_A;
      } else {
        step_pin = STEP_B;
      }
//
      step_forward_timer_started = true;
      stepper_delay_start();        // Starts the stepper motor step rate one-shot timer
      goto go_wait;

//      nrf_delay_ms(STEPPER_MOTOR_RATE); // X mS delay for motor. Change this to a timer
  next_for_step:
      num_steps--;
      step_forward_timer_started = false;
    } while (num_steps != 0);
  go_wait:;
}

      // Step Backward direction
// void step_backward(uint32_t num_steps) {     // eliminate passing parameters so we can use timers
void step_backward(void) {

    if (step_backward_timer_started == true){
//      step_backward_timer_started = false;
      goto next_back_step;
    }
    stepper_A = nrf_gpio_pin_out_read(STEP_A); // Read the current state of STEP_A and STEP_B pin levels
    stepper_B = nrf_gpio_pin_out_read(STEP_B);

    if (stepper_A && stepper_B || !stepper_A && !stepper_B) { // Determine which drive pin (STEP_A or STEP_B) needs to be
      step_pin = STEP_A;                                      // changed first in order to move in the backward direction.
    } else {
      step_pin = STEP_B;
    }
      // Now Step the requested number of steps
    do {
      nrf_gpio_pin_toggle(step_pin); // First step in the backward direction
      if (step_pin == STEP_B) {
        step_pin = STEP_A;
      } else {
        step_pin = STEP_B;
      }
//
      step_backward_timer_started = true;
      stepper_delay_start();        // Starts the stepper motor step rate one-shot timer
      goto go_and_wait;

//        nrf_delay_ms(STEPPER_MOTOR_RATE); // 50mS delay. Change this to timer ticks
  next_back_step:
      num_steps--;
      step_backward_timer_started = false;
    } while (num_steps != 0);
  go_and_wait:;
}

void stepper_delay(void){
  if (step_forward_timer_started == true){
    step_forward();
  } else if (step_backward_timer_started == true) {
    step_backward();
  }
}

/*****************  End of Stepper Motor Timer Code verison *****************************/

/*********************************************************************
*                    End of Stepper Motor Control
**********************************************************************/

/*****************************************************************************************
                Sample ADC test code.  Modified to read system voltage on P0.30/AIN6
******************************************************************************************/
/* This file contains the source code for a sample application using SAADC.  \
// We can't use timer 0 because it is used by the soft-device.  Change timer number. */

int get_sys_v(){

// figure out this code.

}               

#define SAMPLES_IN_BUFFER 5
volatile uint8_t state = 1;

static const nrf_drv_timer_t m_timer = NRF_DRV_TIMER_INSTANCE(1); // Eric - changed from 0 to 1. 0 is used by SD
static nrf_saadc_value_t m_buffer_pool[2][SAMPLES_IN_BUFFER];
static nrf_ppi_channel_t m_ppi_channel;
static uint32_t m_adc_evt_counter;

void timer_handler(nrf_timer_event_t event_type, void *p_context) {
  // nrf_drv_gpiote_out_set(Kontrol_1);
  printf("saadc timer timed out");
  perfect_shower_main();
}

void saadc_sampling_event_init(void) {
  ret_code_t err_code;

  err_code = nrf_drv_ppi_init();
  APP_ERROR_CHECK(err_code);

  nrf_drv_timer_config_t timer_cfg = NRF_DRV_TIMER_DEFAULT_CONFIG;
  timer_cfg.bit_width = NRF_TIMER_BIT_WIDTH_32;
  err_code = nrf_drv_timer_init(&m_timer, &timer_cfg, timer_handler);
  APP_ERROR_CHECK(err_code);

  //    setup m_timer for compare event every 400ms
  uint32_t ticks = nrf_drv_timer_ms_to_ticks(&m_timer, 400);
  nrf_drv_timer_extended_compare(&m_timer,
      NRF_TIMER_CC_CHANNEL0,
      ticks,
      NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK,
      false);
  nrf_drv_timer_enable(&m_timer);

  uint32_t timer_compare_event_addr = nrf_drv_timer_compare_event_address_get(&m_timer,
      NRF_TIMER_CC_CHANNEL0);
  uint32_t saadc_sample_task_addr = nrf_drv_saadc_sample_task_get();

  //    setup ppi channel so that timer compare event is triggering sample task in SAADC
  err_code = nrf_drv_ppi_channel_alloc(&m_ppi_channel);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_ppi_channel_assign(m_ppi_channel,
      timer_compare_event_addr,
      saadc_sample_task_addr);
  APP_ERROR_CHECK(err_code);
}

void saadc_sampling_event_enable(void) {
  ret_code_t err_code = nrf_drv_ppi_channel_enable(m_ppi_channel);

  APP_ERROR_CHECK(err_code);
}

void saadc_callback(nrf_drv_saadc_evt_t const *p_event) {
  if (p_event->type == NRF_DRV_SAADC_EVT_DONE) {
    ret_code_t err_code;

    err_code = nrf_drv_saadc_buffer_convert(p_event->data.done.p_buffer, SAMPLES_IN_BUFFER);
    APP_ERROR_CHECK(err_code);

    int i;
    NRF_LOG_INFO("ADC event number: %d", (int)m_adc_evt_counter);

    for (i = 0; i < SAMPLES_IN_BUFFER; i++) {
      NRF_LOG_INFO("%d", p_event->data.done.p_buffer[i]);
    }
    m_adc_evt_counter++;
  }
}

void saadc_init(void) {
  ret_code_t err_code;
  nrf_saadc_channel_config_t channel_config =
      NRF_DRV_SAADC_DEFAULT_CHANNEL_CONFIG_SE(NRF_SAADC_INPUT_AIN6); // PS uses AIN6 for Sys_Pwr monitoring

  err_code = nrf_drv_saadc_init(NULL, saadc_callback);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_channel_init(0, &channel_config);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[0], SAMPLES_IN_BUFFER);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[1], SAMPLES_IN_BUFFER);
  APP_ERROR_CHECK(err_code);
  // }

  // @brief Function for main application entry.

  // uint32_t get_sys_v(void)     // Get the system (Generator) voltage. Sys_V=1.0v for 18V Gen voltage (nominal), 3.06V for 51V Gen voltage.

  //********************************************************************
  // Moved this code into the main function - DELETE WHEN READY
  //********************************************************************
  /*
    uint32_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();

    ret_code_t ret_code = nrf_pwr_mgmt_init();
    APP_ERROR_CHECK(ret_code);

    saadc_init();
    saadc_sampling_event_init();
*/
  //********************************************************************

  // We only want to sample when we need the voltage info... Not continually

  int gen_voltage; // System voltage is derrived from the generator output voltage, nominally 18V.

  //    saadc_sampling_event_enable();
  NRF_LOG_INFO("SAADC HAL simple example started.");

  //    while (1)
  //    {
  //        nrf_pwr_mgmt_run();
  //        NRF_LOG_FLUSH();
  //    }
  return gen_voltage;
}

//********************************************************************
//                        End SAADC code
//********************************************************************
/*****************************************************************************************
                  MORE ALTERNATIVE FLASH CODE
/******************************************************************************************/

/* A tag identifying the SoftDevice BLE configuration. 
#define APP_BLE_CONN_CFG_TAG 1

/* Array to map FDS return values to strings. 
char const *fds_err_str[] =
    {
        "FDS_SUCCESS",
        "FDS_ERR_OPERATION_TIMEOUT",
        "FDS_ERR_NOT_INITIALIZED",
        "FDS_ERR_UNALIGNED_ADDR",
        "FDS_ERR_INVALID_ARG",
        "FDS_ERR_NULL_ARG",
        "FDS_ERR_NO_OPEN_RECORDS",
        "FDS_ERR_NO_SPACE_IN_FLASH",
        "FDS_ERR_NO_SPACE_IN_QUEUES",
        "FDS_ERR_RECORD_TOO_LARGE",
        "FDS_ERR_NOT_FOUND",
        "FDS_ERR_NO_PAGES",
        "FDS_ERR_USER_LIMIT_REACHED",
        "FDS_ERR_CRC_CHECK_FAILED",
        "FDS_ERR_BUSY",
        "FDS_ERR_INTERNAL",
};

/* Array to map FDS events to strings. 
static char const *fds_evt_str[] =
    {
        "FDS_EVT_INIT",
        "FDS_EVT_WRITE",
        "FDS_EVT_UPDATE",
        "FDS_EVT_DEL_RECORD",
        "FDS_EVT_DEL_FILE",
        "FDS_EVT_GC",
};

/* Dummy configuration data. 
static configuration_t m_dummy_cfg =
    {
        .config1_on = false,
        .config2_on = true,
        .boot_count = 0x0,
        .device_name = "dummy",
};

/* A record containing dummy configuration data. 
static fds_record_t const m_dummy_record =
    {
        .file_id = CONFIG_FILE,
        .key = CONFIG_REC_KEY,
        .data.p_data = &m_dummy_cfg,
        // The length of a record is always expressed in 4-byte units (words). 
        .data.length_words = (sizeof(m_dummy_cfg) + 3) / sizeof(uint32_t),
};

/* Keep track of the progress of a delete_all operation. 
static struct
{
  bool delete_next; //!< Delete next record.
  bool pending;     //!< Waiting for an fds FDS_EVT_DEL_RECORD event, to delete the next record.
} m_delete_all;

/* Flag to check fds initialization. 
static bool volatile m_fds_initialized;

static void fds_evt_handler(fds_evt_t const *p_evt) {
  NRF_LOG_GREEN("Event: %s received (%s)",
      fds_evt_str[p_evt->id],
      fds_err_str[p_evt->result]);

  switch (p_evt->id) {
  case FDS_EVT_INIT:
    if (p_evt->result == FDS_SUCCESS) {
      m_fds_initialized = true;
    }
    break;

  case FDS_EVT_WRITE: {
    if (p_evt->result == FDS_SUCCESS) {
      NRF_LOG_INFO("Record ID:\t0x%04x", p_evt->write.record_id);
      NRF_LOG_INFO("File ID:\t0x%04x", p_evt->write.file_id);
      NRF_LOG_INFO("Record key:\t0x%04x", p_evt->write.record_key);
    }
  } break;

  case FDS_EVT_DEL_RECORD: {
    if (p_evt->result == FDS_SUCCESS) {
      NRF_LOG_INFO("Record ID:\t0x%04x", p_evt->del.record_id);
      NRF_LOG_INFO("File ID:\t0x%04x", p_evt->del.file_id);
      NRF_LOG_INFO("Record key:\t0x%04x", p_evt->del.record_key);
    }
    m_delete_all.pending = false;
  } break;

  default:
    break;
  }
}

/**@brief   Begin deleting all records, one by one. 
void delete_all_begin(void) {
  m_delete_all.delete_next = true;
}

/**@brief   Process a delete all command.
 *
 * Delete records, one by one, until no records are left.
 
void delete_all_process(void) {
  if (m_delete_all.delete_next & !m_delete_all.pending) {
    NRF_LOG_INFO("Deleting next record.");

    m_delete_all.delete_next = record_delete_next();
    if (!m_delete_all.delete_next) {
      NRF_LOG_CYAN("No records left to delete.");
    }
  }
}
*/

/* **************************** THIS CODE ALREADY IMPLEMENTED ******************
#ifdef SOFTDEVICE_PRESENT
/**@brief   Function for initializing the SoftDevice and enabling the BLE stack. 
static void ble_stack_init(void)
{
    ret_code_t rc;
    uint32_t   ram_start;

    /* Enable the SoftDevice. 
    rc = nrf_sdh_enable_request();
    APP_ERROR_CHECK(rc);

    rc = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(rc);

    rc = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(rc);
}
#else
*/
static void clock_init(void) {
  /* Initialize the clock. */
  ret_code_t rc = nrf_drv_clock_init();
  APP_ERROR_CHECK(rc);

  nrf_drv_clock_lfclk_request(NULL);

  /* Wait for the clock to be ready. */
  while (!nrf_clock_lf_is_running()) {
    ;
  }
}

/*
/**@brief   Initialize logging.
static void log_init(void)
{
    ret_code_t rc = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(rc);
}
*/

/**@brief   Sleep until an event is received. */
static void power_manage(void) {
#ifdef SOFTDEVICE_PRESENT
  (void)sd_app_evt_wait();
#else
  __WFE();
#endif
}

/**@brief   Wait for fds to initialize. 
static void wait_for_fds_ready(void) {
  while (!m_fds_initialized) {
    power_manage();
  }
}

int access_flash(void) {
  ret_code_t rc;
  /*
#ifdef SOFTDEVICE_PRESENT
    ble_stack_init();
#else
    clock_init();
#endif

    timer_init();
    log_init();
    cli_init();

    NRF_LOG_INFO("FDS example started.")

    /* Register first to receive an event when initialization is complete. 
  (void)fds_register(fds_evt_handler);

  NRF_LOG_INFO("Initializing fds...");

  rc = fds_init();
  APP_ERROR_CHECK(rc);

  /* Wait for fds to initialize. 
  wait_for_fds_ready();

  NRF_LOG_INFO("Available commands:");
  NRF_LOG_INFO("- print all\t\tprint records");
  NRF_LOG_INFO("- print config\tprint configuration");
  NRF_LOG_INFO("- update\t\tupdate configuration");
  NRF_LOG_INFO("- stat\t\tshow statistics");
  NRF_LOG_INFO("- write\t\twrite a new record");
  NRF_LOG_INFO("- delete\t\tdelete a record");
  NRF_LOG_INFO("- delete_all\tdelete all records");
  NRF_LOG_INFO("- gc\t\trun garbage collection");

  NRF_LOG_INFO("Reading flash usage statistics...");

  fds_stat_t stat = {0};

  rc = fds_stat(&stat);
  APP_ERROR_CHECK(rc);

  NRF_LOG_INFO("Found %d valid records.", stat.valid_records);
  NRF_LOG_INFO("Found %d dirty records (ready to be garbage collected).", stat.dirty_records);

  fds_record_desc_t desc = {0};
  fds_find_token_t tok = {0};

  rc = fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok);

  if (rc == FDS_SUCCESS) {
    /* A config file is in flash. Let's update it. 
    fds_flash_record_t config = {0};

    /* Open the record and read its contents. 
    rc = fds_record_open(&desc, &config);
    APP_ERROR_CHECK(rc);

    /* Copy the configuration from flash into m_dummy_cfg. 
    memcpy(&m_dummy_cfg, config.p_data, sizeof(configuration_t));

    NRF_LOG_INFO("Config file found, updating boot count to %d.", m_dummy_cfg.boot_count);

    /* Update boot count. 
    m_dummy_cfg.boot_count++;

    /* Close the record when done reading. 
    rc = fds_record_close(&desc);
    APP_ERROR_CHECK(rc);

    /* Write the updated record to flash. 
    rc = fds_record_update(&desc, &m_dummy_record);
    APP_ERROR_CHECK(rc);
  } else {
    /* System config not found; write a new one. 
    NRF_LOG_INFO("Writing config file...");

    rc = fds_record_write(&desc, &m_dummy_record);
    APP_ERROR_CHECK(rc);
  }

  cli_start();

  /* Enter main loop. 
  for (;;) {
    if (!NRF_LOG_PROCESS()) {
      power_manage();
    }
    cli_process();
    delete_all_process();
  }
}
*/

/**********************************************************************
************* PWM code for Audio out  *********************************
***********************************************************************/
APP_PWM_INSTANCE(PWM1,2);                   // Create the instance "PWM1" using TIMER 2.

static volatile bool ready_flag;            // A flag indicating PWM status.

void pwm_ready_callback(uint32_t pwm_id)    // PWM callback function
{
    ready_flag = true;
}

int pwm_init(void)
{
    ret_code_t err_code;

    // 1-channel PWM 
    app_pwm_config_t pwm1_cfg = APP_PWM_DEFAULT_CONFIG_1CH(850L, AUDIO_OUT);

    // Initialize and enable PWM. 
    err_code = app_pwm_init(&PWM1,&pwm1_cfg,pwm_ready_callback);
    APP_ERROR_CHECK(err_code);
    app_pwm_enable(&PWM1);

    uint32_t pwm_value;
      pwm_value = 80;
      ready_flag = false;
            // Set the duty cycle - keep trying until PWM is ready... 
      while (app_pwm_channel_duty_set(&PWM1, 0, pwm_value) == NRF_ERROR_BUSY);

            // ... or wait for callback. 
      while (!ready_flag);
      APP_ERROR_CHECK(app_pwm_channel_duty_set(&PWM1, 1, pwm_value));
//      nrf_delay_ms(25);
}

/*
void pwm_init(void);
{
    APP_PWM_INSTANCE(PWM1,2);                   // Create the instance "PWM1" using TIMER2.
    static volatile bool ready_flag;            // A flag indicating PWM status.
}
void pwm_ready_callback(uint32_t pwm_id)    // PWM callback function
{
    ready_flag = true;
}
int pwm_run(void)
{
    ret_code_t err_code;

    /* 1-channel PWM, 200Hz, output on AUDIO_OUT pin. 
    app_pwm_config_t pwm1_cfg = APP_PWM_DEFAULT_CONFIG_1CH(5000L, AUDIO_OUT);

    /* Initialize and enable PWM. 
    err_code = app_pwm_init(&PWM1,&pwm1_cfg,pwm_ready_callback);
    APP_ERROR_CHECK(err_code);
    app_pwm_enable(&PWM1);

    uint32_t value;
    while (true)
    {
        for (uint8_t i = 0; i < 40; ++i)
        {
            value = (i < 20) ? (i * 5) : (100 - (i - 20) * 5);

            ready_flag = false;
            /* Set the duty cycle - keep trying until PWM is ready... 
            while (app_pwm_channel_duty_set(&PWM1, 0, value) == NRF_ERROR_BUSY);

            /* ... or wait for callback. 
            while (!ready_flag);
            APP_ERROR_CHECK(app_pwm_channel_duty_set(&PWM1, 1, value));
            nrf_delay_ms(25);
        }
    }

}

/*
static nrf_drv_pwm_t m_pwm0 = NRF_DRV_PWM_INSTANCE(0);
static uint8_t m_used = USED_PWM(0);
#define USED_PWM(idx) (1UL << idx)
static void audio_tone_output(void)
{
    /*
     * This demo uses only one channel, which is reflected on LED 1.
     * The LED blinks three times (200 ms on, 200 ms off), then it stays off
     * for one second.
     * This scheme is performed three times before the peripheral is stopped.
     */
/*
    nrf_drv_pwm_config_t const config0 =
    {
        .output_pins = 
        {
            AUDIO_OUT | NRF_DRV_PWM_PIN_INVERTED, // channel 0
        },
        .irq_priority = APP_IRQ_PRIORITY_LOWEST,
        .base_clock   = NRF_PWM_CLK_125kHz,
        .count_mode   = NRF_PWM_MODE_UP,
        .top_value    = 25000,
        .load_mode    = NRF_PWM_LOAD_COMMON,
        .step_mode    = NRF_PWM_STEP_AUTO
    };
    APP_ERROR_CHECK(nrf_drv_pwm_init(&m_pwm0, &config0, NULL));
//    m_used |= USED_PWM(0);

    // This array cannot be allocated on stack (hence "static") and it must
    // be in RAM (hence no "const", though its content is not changed).
    static uint16_t seq_values[] =
 /*   {
        0x8000,
             0,
        0x8000,
             0,
        0x8000,
             0
    };
    nrf_pwm_sequence_t const seq =
    {
        .values.p_common = seq_values,
        .length          = NRF_PWM_VALUES_LENGTH(seq_values),
        .repeats         = 1,   // Eric was 0
        .end_delay       = 4
    };

    (void)nrf_drv_pwm_simple_playback(&m_pwm0, &seq, 3, NRF_DRV_PWM_FLAG_STOP);
}
*/

/**********************************************************************
************* Eric's Original Arduino Code & New Nordic Code  *********
***********************************************************************

/*****************  Manual Factory Reset  ********************
//  Read Reset pin GPIO 5 to see if we need to perform a system reset
    int pinState = digitalRead(system_Voltage);  // Need to change to digital input vs analog?
  if (pinState == 0){                            // Set all of the flags to reset system.
    initializationDone = false;
    install_mode = true;
  }

/*******************************  WARNING/DANGER  ***********************************
 *   Make sure that the WG_Temp can never be higher than the absoluteMaxTemp variable
 ************************************************************************************

/*****************************************************************************************
                END - MORE ALTERNATIVE FLASH CODE
/******************************************************************************************/

/*****************************************************************************************
                Flash Write Example code to modify for PS
******************************************************************************************/

// @brief This file contains the source code for a sample application using the Flash Write Application.
/* 

#define FLASHWRITE_EXAMPLE_MAX_STRING_LEN       (62u)
#define FLASHWRITE_EXAMPLE_BLOCK_VALID          (0xA55A5AA5)
#define FLASHWRITE_EXAMPLE_BLOCK_INVALID        (0xA55A0000)
#define FLASHWRITE_EXAMPLE_BLOCK_NOT_INIT       (0xFFFFFFFF)

NRF_CLI_UART_DEF(m_cli_uart_transport, 0, 64, 16);
NRF_CLI_DEF(m_cli_uart, "uart_cli:~$ ", &m_cli_uart_transport.transport, '\r', 4);
NRF_MODULE_ENABLED(NRF_CLI)

typedef struct
{
   uint32_t magic_number;
   uint32_t buffer[FLASHWRITE_EXAMPLE_MAX_STRING_LEN + 1]; // + 1 for end of string
} flashwrite_example_flash_data_t;

typedef struct
{
    uint32_t addr;
    uint32_t pg_size;
    uint32_t pg_num;
    flashwrite_example_flash_data_t * m_p_flash_data;
} flashwrite_example_data_t;

static flashwrite_example_data_t m_data;

static ret_code_t clock_config(void)
{
    ret_code_t err_code;

    err_code = nrf_drv_clock_init();
    if (err_code != NRF_SUCCESS && err_code != NRF_ERROR_MODULE_ALREADY_INITIALIZED)
    {
        return err_code;
    }

    nrf_drv_clock_lfclk_request(NULL);

    return NRF_SUCCESS;
}

static void flash_page_init(void)
{
    m_data.pg_num = NRF_FICR->CODESIZE - 1;
    m_data.pg_size = NRF_FICR->CODEPAGESIZE;
    m_data.addr = (m_data.pg_num * m_data.pg_size);

    m_data.m_p_flash_data = (flashwrite_example_flash_data_t *)m_data.addr;

    while (1)
    {
        if (m_data.m_p_flash_data->magic_number == FLASHWRITE_EXAMPLE_BLOCK_VALID)
        {
            return;
        }

        if (m_data.m_p_flash_data->magic_number == FLASHWRITE_EXAMPLE_BLOCK_INVALID)
        {
            ++m_data.m_p_flash_data;
            continue;
        }

        nrf_nvmc_page_erase(m_data.addr);
        return;
    }
}

// @brief Function for application main entry.
 
int write_flash(void)
{
    uint32_t err_code;

    APP_ERROR_CHECK(NRF_LOG_INIT(NULL));

    err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    err_code = clock_config();
    APP_ERROR_CHECK(err_code);

    flash_page_init();

    nrf_drv_uart_config_t uart_config = NRF_DRV_UART_DEFAULT_CONFIG;
    uart_config.pseltxd = TX_PIN_NUMBER;
    uart_config.pselrxd = RX_PIN_NUMBER;
    uart_config.hwfc    = NRF_UART_HWFC_DISABLED;
    err_code = nrf_cli_init(&m_cli_uart, &uart_config, true, true, NRF_LOG_SEVERITY_INFO);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_cli_start(&m_cli_uart);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_RAW_INFO("Flashwrite example started.\r\n");
    NRF_LOG_RAW_INFO("Execute: <flash -h> for more information "
                     "or press the Tab button to see all available commands.\r\n");

    while (true)
    {
        UNUSED_RETURN_VALUE(NRF_LOG_PROCESS());
        nrf_cli_process(&m_cli_uart);
    }
}

static void flash_string_write(uint32_t address, const char * src, uint32_t num_words)
{
    uint32_t i;

    // Enable write.
    NRF_NVMC->CONFIG = NVMC_CONFIG_WEN_Wen;
    while (NRF_NVMC->READY == NVMC_READY_READY_Busy)
    {
    }

    for (i = 0; i < num_words; i++)
    {
        /* Only full 32-bit words can be written to Flash. 
        ((uint32_t*)address)[i] = 0x000000FFUL & (uint32_t)((uint8_t)src[i]);
        while (NRF_NVMC->READY == NVMC_READY_READY_Busy)
        {
        }
    }

    NRF_NVMC->CONFIG = NVMC_CONFIG_WEN_Ren;
    while (NRF_NVMC->READY == NVMC_READY_READY_Busy)
    {
    }
}


static void flashwrite_erase_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    nrf_nvmc_page_erase(m_data.addr);

    m_data.m_p_flash_data = (flashwrite_example_flash_data_t *)m_data.addr;
}

static void flashwrite_read_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    flashwrite_example_flash_data_t * p_data = (flashwrite_example_flash_data_t *)m_data.addr;
    char string_buff[FLASHWRITE_EXAMPLE_MAX_STRING_LEN + 1]; // + 1 for end of string

    if ((p_data == m_data.m_p_flash_data) &&
        (p_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_VALID))
    {
        nrf_cli_fprintf(p_cli, NRF_CLI_WARNING, "Please write something first.\r\n");
        return;
    }

    while (p_data <= m_data.m_p_flash_data)
    {
        if ((p_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_VALID) &&
            (p_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_INVALID))
        {
            nrf_cli_fprintf(p_cli, NRF_CLI_WARNING, "Corrupted data found.\r\n");
            return;
        }
        uint8_t i;
        for (i = 0 ; i <= FLASHWRITE_EXAMPLE_MAX_STRING_LEN; i++)
        {
            string_buff[i] = (char)p_data->buffer[i];
        }

        nrf_cli_fprintf(p_cli, NRF_CLI_NORMAL, "%s\r\n", string_buff);
        ++p_data;
    }
}

static void flashwrite_write_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    static uint16_t const page_size = 4096;

    if (argc < 2)
    {
        nrf_cli_fprintf(p_cli, NRF_CLI_ERROR, "%s:%s", argv[0], " bad parameter count\r\n");
        return;
    }
    if (argc > 2)
    {
        nrf_cli_fprintf(p_cli,
                        NRF_CLI_WARNING,
                        "%s:%s",
                        argv[0],
                        " bad parameter count - please use quotes\r\n");
        return;
    }

    uint32_t len = strlen(argv[1]);
    if (len > FLASHWRITE_EXAMPLE_MAX_STRING_LEN)
    {
        nrf_cli_fprintf(p_cli,
                        NRF_CLI_ERROR,
                        "Too long string. Please limit entered string to %d chars.\r\n",
                        FLASHWRITE_EXAMPLE_MAX_STRING_LEN);
        return;
    }

    if ((m_data.m_p_flash_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_NOT_INIT) &&
        (m_data.m_p_flash_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_VALID))
    {
        nrf_cli_fprintf(p_cli, NRF_CLI_ERROR, "Flash corrupted, please errase it first.");
        return;
    }

    if (m_data.m_p_flash_data->magic_number == FLASHWRITE_EXAMPLE_BLOCK_VALID)
    {
        uint32_t new_end_addr = (uint32_t)(m_data.m_p_flash_data + 2);
        uint32_t diff = new_end_addr - m_data.addr;
        if (diff > page_size)
        {
            nrf_cli_fprintf(p_cli,
                            NRF_CLI_WARNING,
                            "Not enough space - please erase flash first.\r\n");
            return;
        }
        nrf_nvmc_write_word((uint32_t)&m_data.m_p_flash_data->magic_number,
                            FLASHWRITE_EXAMPLE_BLOCK_INVALID);
        ++m_data.m_p_flash_data;
    }

    //++len -> store also end of string '\0'
    flash_string_write((uint32_t)m_data.m_p_flash_data->buffer, argv[1], ++len);
    nrf_nvmc_write_word((uint32_t)&m_data.m_p_flash_data->magic_number,
                        FLASHWRITE_EXAMPLE_BLOCK_VALID);
}

static void flashwrite_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    ASSERT(p_cli);
    ASSERT(p_cli->p_ctx && p_cli->p_iface && p_cli->p_name);

    if ((argc == 1) || nrf_cli_help_requested(p_cli))
    {
        nrf_cli_help_print(p_cli, NULL, 0);
        return;
    }

    nrf_cli_fprintf(p_cli, NRF_CLI_ERROR, "%s:%s%s\r\n", argv[0], " unknown parameter: ", argv[1]);
}

NRF_CLI_CREATE_STATIC_SUBCMD_SET(m_sub_flash)
{
    NRF_CLI_CMD(erase, NULL, "Erase flash.",          flashwrite_erase_cmd),
    NRF_CLI_CMD(read,  NULL, "Read data from flash.", flashwrite_read_cmd),
    NRF_CLI_CMD(write, NULL, "Write data to flash.\n"
                             "Limitations:\n"
                             "- maximum 16 entries,\n"
                             "- each entry is maximum 62 chars long.",
                                                      flashwrite_write_cmd),
    NRF_CLI_SUBCMD_SET_END
};
NRF_CLI_CMD_REGISTER(flash, &m_sub_flash, "Flash access command.", flashwrite_cmd);
*/
/*****************************************************************************************
                End - Flash Write Example code to modify for PS
******************************************************************************************


/*****************************************************************************************
                          ALTERNATIVE FLASH CODE
******************************************************************************************/
/*

static void fstorage_test(void)
{
		static uint32_t data;
		uint32_t flash_data[4];
	
		FS_REGISTER_CFG(fs_config_t fs_config) =
		{
			  .callback  = fs_evt_handler, // Function for event callbacks.
				.num_pages = NUM_PAGES,      // Number of physical flash pages required.
				.priority  = 0xFE            // Priority for flash usage.
		};
		
		fs_ret_t ret = fs_init();
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		
		// Erase one page (page 0).
		printf("Erasing a flash page at address 0x%X\r\n", (uint32_t)fs_config.p_start_addr);
		fs_callback_flag = 1;
		ret = fs_erase(&fs_config, fs_config.p_start_addr, 1);
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		while(fs_callback_flag == 1)  { power_manage(); }
		
		//Read the first 4 words of the page
		printf("Data read from flash address 0x%X: ", (uint32_t)fs_config.p_start_addr);
		for(int i=0; i<4; i++)
		{
				flash_data[i] = *(fs_config.p_start_addr + i);
				printf("%X ", flash_data[i]);
		}
		printf("\r\n");
		
		//Write the first byte
		data = 0xAAAAAAAA;
		printf("Writing data 0x%X to address 0x%X\r\n", data, (uint32_t)fs_config.p_start_addr);
		fs_callback_flag = 1;
		ret = fs_store(&fs_config, fs_config.p_start_addr, &data, 1);      //Write data to memory address 0x0003F000. Check it with command: nrfjprog --memrd 0x0003F000 --n 16
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		while(fs_callback_flag == 1)  { power_manage(); }
		
		//Write second byte
		data = 0xBBBBBBBB;
		printf("Writing data 0x%X to address 0x%X\r\n", data, (uint32_t)fs_config.p_start_addr + 4);
		fs_callback_flag = 1;
		ret = fs_store(&fs_config, fs_config.p_start_addr + 1, &data, 1);      //Write data to memory address 0x0003F000. Check it with command: nrfjprog --memrd 0x0003F000 --n 16
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		while(fs_callback_flag == 1)  { power_manage(); }
		
		//Read the first 4 words of the page
		printf("Data read from flash address 0x%X: ", (uint32_t)fs_config.p_start_addr);
		for(int i=0; i<4; i++)
		{
				flash_data[i] = *(fs_config.p_start_addr + i);
				printf("%X ", flash_data[i]);
		}
		printf("\r\n");
}


/*****************************************************************************************
                        END - ALTERNATIVE FLASH CODE
******************************************************************************************/
/*
/*****************************************************************************************
                        ANOTHER FLASH WRITE EXAMPLE
******************************************************************************************/
/*
fds_init();
nrf_delay_ms(300);

__ALIGN(4) beacon_packet_struct   beacon_flags;

typedef struct
{
    uint8_t led_blink; //was bool
    uint8_t initial_adv_disable; //was bool
    uint8_t enable_ibeacon_Packet; //was bool
    uint8_t enable_cbeacon1_Health_Packet; //was bool
    uint8_t enable_cbeacon2_Proximity_Packet;  //was bool
    uint8_t enable_ibeacon_SOS_Packet; //was bool
    uint8_t enable_Mems; //was bool
}beacon_packet_struct;


int fds_write_struct(uint16_t file_id, uint16_t flash_key)
{
    ret_code_t ret;

    fds_record_t        record;
    fds_record_desc_t   record_desc;

    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

	if (flash_key == BEACON_STATE_KEY)
    {

        record.file_id              = FILE_ID;
        record.key                  = BEACON_STATE_KEY;
        record.data.p_data          = &beacon_flags;  //&beacon_flags;
        record.data.length_words    = ( sizeof(beacon_packet_struct) + 3 ) / sizeof(uint32_t);
      
        ret_code_t rc = fds_record_find(FILE_ID, BEACON_STATE_KEY , &desc, &tok);  //Search for record. Update if found else write a new one

        if (rc == FDS_SUCCESS)
        {   
            NRF_LOG_INFO("Similar BEACON STATE Variables Record was found thus it is being updated");
            rc = fds_record_update(&desc, &record);
            APP_ERROR_CHECK(rc);
        }
        else
        {
            ret_code_t ret = fds_record_write(&record_desc, &record);
            APP_ERROR_CHECK(ret);
            NRF_LOG_INFO("First time writing BEACON STATE record");
        }

        return 1;
    }
	
}
/*****************************************************************************************
                      END- ANOTHER FLASH WRITE EXAMPLE
******************************************************************************************/